#include "All.h"
#include "machine.h"
#include "sound.h"
#include "myprocess.h"
#include "debug.h"

	extern	unsigned char	BreakPoints[65536];		// break on....blah
	extern	int		debugger;
	extern	int		pROM[];				// plus/4 ROM
	extern	int		ROM3PLUS1[];				// plus/4 ROM
	extern	int		Joy1Enable;			// Joystick 1 off by default
	extern	int		Joy2Enable;			// Joystick 2 on  by default

#define	SOUND__DELAY			(20)

#define	final
	// 8501 flags
#define	__C	1
#define	__Z  2
#define	__I	4
#define	__D	8
#define	__B	16
#define	__V	64
#define	__N	128

	int		Keys[13];
	int		skipit;

	int		CompTable[256];				// plus/4 memory	

	int		__A;							// registers
	int		__X;
	int		__Y;
	int		__SP;
	int		__F;
	int		__PC;
	
	int		Memory[65536];				// plus/4 memory
	int		*pMemory = &Memory[0];

	// Profiling. Read/Write/Execute.
	int		ReadProfile[65536];
	int		WriteProfile[65536];
	int		ExecuteProfile[65536];
	int		MaxRead;
	int		MaxWrite;
	int		MaxExecute;

	int		ROM1[16384];
	int		ROM2[16384];
	int		ROM3[16384];
	int		ROM4[16384];
	int		ROM5[16384];
	int		ROM6[16384];
	int		ROM7[16384];
	int		ROM8[16384];
	int		*pRomLo = &pROM[0];			// Basic
	int		*pRomHi = &pROM[16384];		// Kernal


	int		*pRom0 = &pROM[0];
	int		*pRom1 = &ROM3PLUS1[0];
	int		*pRom2 = &pROM[0];
	int		*pRom3 = &pROM[0];
	int		*pRom4 = &pROM[16384];
	int		*pRom5 = &ROM3PLUS1[16384];
	int		*pRom6 = &pROM[0];
	int		*pRom7 = &pROM[0];

	int		g_TimerDelay1 = 0;
	int		Cycles;
	int		opcode;

	int		data;
	int		d1;

	int		ROMACTIVE;
	int		breakpoint=0xdb11;


extern	int		Freq1Timer;
extern	int		Freq2Timer;	
extern	int		ChannelControlTimer;

//************************************************************
//                                                           *
//   Name    : MakeCompTable                                 *
//   Function: this creates an array for comparing a register*
//             and what flags we get back from it.           *
//                                                           *
//************************************************************

void MakeCompTable( )
{
	int	F;
	int	test;
	
	for (test=0;test<256;test++){
		F=0;
		if ((test & 128)==128) {F|=__N;}
		if (test==0) {F|=__Z;}
		CompTable[test]=F;
	}
}

void PatternFillMemory( void );
void Init6502( )
{
	int	i;
//	for (i=0;i<65546;i++){
//		pMemory[i]=0;
//	}
	
	PatternFillMemory();				// fill memory with an ff ff 00 00 pattern (like the plus4)

	// Clear HW registers
	for (i=0xff00;i<0x10000;i++){
		pMemory[i]=0;

		ReadProfile[i] = 0;
		WriteProfile[i] = 0;
		ExecuteProfile[i] = 0;
	}

	MaxRead = 0;
	MaxWrite = 0;
	MaxExecute = 0;
	skipit=0;
	ROMACTIVE=1;
	Cycles=0;
	opcode=0;
	__A=0;
	__X=0;
	__Y=0;
	__SP=0x1ff;
	__F=0;
	__PC=0;
	MakeCompTable();
	Keys[0]=Keys[1]=Keys[2]=Keys[3]=Keys[4]=Keys[5]=Keys[6]=Keys[7]=Keys[8]=Keys[9]=Keys[10]=0xff;

}

//****************************************************************************
// Read the keybaord 
//
// a = value written into $FD30
//****************************************************************************
int ReadKeyboard( int a )
{
	int	key = pMemory[0x0fd30];
	int	b=0xff;

	if ((key&1)==0) b&=Keys[1];
	if ((key&2)==0) b&=Keys[2];
	if ((key&4)==0) b&=Keys[3];
	if ((key&8)==0) b&=Keys[4];
	if ((key&16)==0) b&=Keys[5];
	if ((key&32)==0) b&=Keys[6];
	if ((key&64)==0) b&=Keys[7];
	if ((key&128)==0) b&=Keys[8];

	// Scan joysticks
	if( Joy1Enable == 1 ){
		if( (a & 4 ) == 0 ){
			b&=Keys[9];
		}
	}
	if( Joy2Enable == 1 ){
		if( (a & 2 ) == 0 ){
			b&=Keys[10];
		}
	}
	return b;
}


//****************************************************************************
// COMP - compare data to 6502 flags
//****************************************************************************
void COMP (int data)
{
	__F=(__F&(__I+__D+__V+__C+__B)) | CompTable[data&0xff];
}

//****************************************************************************
// PEEK memory
//****************************************************************************
int Peek(int address)
{
	// Flag debugger to take over.
//	if( (BreakPoints[address]&4)!=0){
//		debugger=1; 
//	}
	address&=0xffff;

	ReadProfile[address]++;
	if( MaxRead<ReadProfile[address]) MaxRead = ReadProfile[address];

	if((pMemory[0xfd21]&0xc0) ==0 ){
		pMemory[0xfd21] ^= 0xc0;
	}
	// IS rom active? if not... read from RAM
	if ((ROMACTIVE==0) || (address<0x8000)) {
		return pMemory[address];
	}else{
		// ALL hardware and TED access via MEMORY
		if( ( address >=0xfd10 ) && (address < 0xff40 ) ){
			switch( address ){
				case	0xfef1: { 
									int	a;
									a = pMemory[ 0xfef1 ];
									pMemory[ 0xfef1 ] = 0;
									return a;
								}			
				case	0xff15: 
				case	0xff16: 
				case	0xff17: 
				case	0xff18: 
				case	0xff19: return pMemory[address] | 0x80; break;

				default: return pMemory[ address ];
			}
		}
		else if( ( address>=0xfc00 ) && ( address<=0xfcff ) ){		// this ALWAYS comes fom main rom.
			return pROM[address&0x7fff];
		}
		else if( address<0xc000 ){
			return pRomLo[address&0x3fff];
		}
		else{
			return pRomHi[address&0x3fff];
		}
	}
}

//****************************************************************************
// DPEEK memory
//****************************************************************************
int Deek(int address)
{
	int	a;
	a = Peek(address);
	a |= (Peek(address+1)<<8);
	return a;
}

//****************************************************************************
// Poke memory - stores into memory, but does NOT process hardware.
//****************************************************************************
void POKE (int address,int val)
{
	address&=0xffff;
	val&=0xff;

	pMemory[address]=val;
	if ((address>=0xff00) && (address<=0xff40)) {
		pROM[address&0x7fff]=val;
	}
}


void DOKE (int address,int val)
{
	POKE (address,val&0xff);
	POKE (address+1,(val>>8)&0xff);
}

//****************************************************************************
// pull data from stack
//****************************************************************************
int	Pull ( )
{
	__SP++;
	if (__SP>0x1ff) __SP-=0x100;
	//return (pMemory[__SP]&0xff);
	return Peek(__SP);
}

//****************************************************************************
// pull data from stack
//****************************************************************************
int Dull ( )
{
	int	a;

	__SP++;
	if (__SP>0x1ff) __SP-=0x100;
	a = Peek( __SP );
		
	__SP++;
	if (__SP>0x1ff) __SP-=0x100;
	a += ( Peek( __SP )<<8);

	return a;
}

//****************************************************************************
// push data into the stack
//****************************************************************************
void PushReg (int reg)
{
	POKE( __SP, reg&0xff );
	__SP--;
	if (__SP<0x100) __SP=0x1ff;
}


//****************************************************************************
// push data into the stack
//****************************************************************************
void DushReg (int reg)
{
	POKE( __SP, (reg>>8)&0xff );
	__SP--;
	if (__SP<0x100) __SP=0x1ff;
	POKE( __SP, reg&0xff );
	__SP--;
	if (__SP<0x100) __SP=0x1ff;
}

//****************************************************************************
// Take a BRANCH instruction - signed offset byte is passed
//****************************************************************************
void TakeBranch( int offset )
{
   if (offset>127) {
	  offset=(-offset)&0xff;
	  __PC=(__PC-offset)&0xffff;
   } else {
	  __PC=(__PC+offset)&0xffff;
   }
}


//****************************************************************************
// Small Hardware functions
//****************************************************************************
// Timer 1
void inline SetTimer1LO(int val)
{
	pMemory[0xff00]=pMemory[0xff01]=0;		// clear timer for CPU reads
//	TIMER1__ACTIVE=255;						// disable timer 1
	Timer1Value=((unsigned int)val)<<16;					// set LO of timer 1
}
void inline SetTimer1(int val)
{
//	TIMER1__ACTIVE=0; 
	Timer1=Timer1Value=((Timer1Value&0xff0000)|((unsigned int)val)<<24); 
	pMemory[0xff00]=(int) ((Timer1>>16)&0xff);    // write NEW timer LOW
	pMemory[0xff01]=(int) ((Timer1>>24)&0xff);    // write NEW timer HIGH
	g_TimerDelay1 = 0;
}

// Timer 2
void inline SetTimer2LO(int val)
{
	pMemory[0xff02]=pMemory[0xff03]=0;		// clear timer for CPU reads
//	TIMER2__ACTIVE=255;						// disable timer 2
	Timer2=((unsigned int)val)<<16;							// set LO of timer 2
}
void inline SetTimer2(int val)
{
//	TIMER2__ACTIVE=0; 
	Timer2=Timer1Value=((Timer2&0xff0000)|((unsigned int)val)<<24); 
	pMemory[0xff02]=(int) ((Timer2>>16)&0xff);    // write NEW timer LOW
	pMemory[0xff03]=(int) ((Timer2>>24)&0xff);    // write NEW timer HIGH
}

// Timer 3
void inline SetTimer3LO(int val)
{
	pMemory[0xff04]=pMemory[0xff05]=0;		// clear timer for CPU reads
//	TIMER3__ACTIVE=255;						// disable timer 3
	Timer3=((unsigned int)val)<<16;							// set LO of timer 3
}
void inline SetTimer3(int val)
{
//	TIMER3__ACTIVE=0; 
	Timer3=Timer1Value=((Timer3&0xff0000)|((unsigned int)val)<<24); 
	pMemory[0xff04]=(int) ((Timer3>>16)&0xff);    // write NEW timer LOW
	pMemory[0xff05]=(int) ((Timer3>>24)&0xff);    // write NEW timer HIGH
}

//****************************************************************************
// Poke memory
//****************************************************************************
#define	SIDMIRROR		(0xfe80-0xfd40)
void Poke(int address,int val)
{
	address&=0xffff;
	val&=0xff;

	WriteProfile[address]++;
	if( MaxWrite<WriteProfile[address]) MaxWrite = WriteProfile[address];

	// Flag debugger to take over.
	if( (BreakPoints[address]&4)!=0){
		debugger=1; 
		WindowPC = __PC;
		CallStackPointer=-1;
	}



//	if(address==0xfd00){
//		address&=0xffff;
//	}

	pMemory[address]=val;
	if((address>=0xfe80) && (address<0xfe99)){
		address -= (0xfe80-0xfd40);
	}
	if((address>=0xfdd0) && (address<=0xfddf)){
		switch(address){
			case	0xfdd0:	pRomLo = pRom0;	pRomHi = pRom4; return;
			case	0xfdd1:	pRomLo = pRom1;	pRomHi = pRom4; return;
			case	0xfdd2:	pRomLo = pRom2;	pRomHi = pRom4; return;
			case	0xfdd3:	pRomLo = pRom3;	pRomHi = pRom4; return;
			case	0xfdd4:	pRomLo = pRom0;	pRomHi = pRom5; return;
			case	0xfdd5:	pRomLo = pRom1;	pRomHi = pRom5; return;
			case	0xfdd6:	pRomLo = pRom2;	pRomHi = pRom5; return;
			case	0xfdd7:	pRomLo = pRom3;	pRomHi = pRom5; return;
			case	0xfdd8:	pRomLo = pRom0;	pRomHi = pRom6; return;
			case	0xfdd9:	pRomLo = pRom1;	pRomHi = pRom6; return;
			case	0xfdda:	pRomLo = pRom2;	pRomHi = pRom6; return;
			case	0xfddb:	pRomLo = pRom3;	pRomHi = pRom6; return;
			case	0xfddc:	pRomLo = pRom0;	pRomHi = pRom7; return;
			case	0xfddd:	pRomLo = pRom1;	pRomHi = pRom7; return;
			case	0xfdde:	pRomLo = pRom2;	pRomHi = pRom7; return;
			case	0xfddf:	pRomLo = pRom3;	pRomHi = pRom7; return;
		}
		return;
	}
	else if((address>=0xfd20) && (address<0xfd40)) {
		switch( address ){
			case	0xfd20:	pMemory[0xfd20] = val; break;
			case	0xfd21:	pMemory[0xfd21] = 0xc0|val; break;
			case	0xfd22:	pMemory[0xfd22] = val; pMemory[0xfd21] = 0; break;
			case	0xfd30:
			case	0xfd31:
			case	0xfd32:
			case	0xfd33:
			case	0xfd34:
			case	0xfd35:
			case	0xfd36:
			case	0xfd37:
			case	0xfd38:
			case	0xfd39:
			case	0xfd3a:
			case	0xfd3b:
			case	0xfd3c:
			case	0xfd3d:
			case	0xfd3e:
			case	0xfd3f:	pMemory[0xfd30] = val; pMemory[address] = val;
			default:
				pMemory[address] = val;
		}
	}
	else if((address>=0xfd40) && (address<0xfd59)) {

		switch (address){
			// Channel 1
			case	0xfd40:	
			case	0xfd41:	
			case	0xfd42:	
			case	0xfd43:	
			case	0xfd44:	
			case	0xfd45:	
			case	0xfd46:	pRomHi[address&0x3fff] = (pMemory[address+SIDMIRROR] = (pMemory[address]=val)); 
							InitChannel1(address,val);
							break;

			// Channel 2
			case	0xfd47:	
			case	0xfd48:	
			case	0xfd49:	
			case	0xfd4a:	
			case	0xfd4b:	
			case	0xfd4c:	
			case	0xfd4d:	pRomHi[address&0x3fff] = (pMemory[address+SIDMIRROR] = (pMemory[address]=val)); 
							InitChannel2(address,val);
							break;				
			
			// Channel 3
			case	0xfd4e:	
			case	0xfd4f:	
			case	0xfd50:	
			case	0xfd51:	
			case	0xfd52:	
			case	0xfd53:	
			case	0xfd54:	pRomHi[address&0x3fff] = (pMemory[address+SIDMIRROR] = (pMemory[address]=val)); 
							InitChannel3(address,val);
							break;				


			// Just store case....Control registers, processed each "Click"
			case	0xfd55:	
			case	0xfd56:	
			case	0xfd57:	
			case	0xfd58:	pRomHi[address&0x3fff] = (pMemory[address+SIDMIRROR] = (pMemory[address]=val)); 
							WriteChannelData(address,val);			
							break;
		}
	}
	else if((address>=0xfef0) && (address<0xff40)) {

		//romaddress=address&0x7fff;
		//pROM[romaddress]=val;

		switch (address){
			// Timers
			case	0xfef0:	ProcessDiskDrive ( val ); break;
			case	0xff00:	pMemory[address]=val; SetTimer1LO(val); break;
			case	0xff01:	pMemory[address]=val; SetTimer1(val);	break;
			case	0xff02:	pMemory[address]=val; SetTimer2LO(val); break;
			case	0xff03:	pMemory[address]=val; SetTimer2(val);	break;
			case	0xff04:	pMemory[address]=val; SetTimer3LO(val); break;
			case	0xff05:	pMemory[address]=val; SetTimer3(val);	break;

			case	0xff08: pROM[0x7f08]=(pMemory[0xff08]=ReadKeyboard(val)); break; 
			case	0xff09: pROM[0x7f09]=(pMemory[0xff09]&=~val); break; 

			case	0xff0f:
			case	0xff10:	Freq2Timer = SOUND__DELAY;
							pMemory[address]=val;
							pROM[address&0x7fff]=val;		// never 
							break;
			case	0xff0e:
			case	0xff12:	Freq1Timer = SOUND__DELAY;
							pMemory[address]=val;
							pROM[address&0x7fff]=val;		// never 
							break;
			case	0xff11:
							ChannelControlTimer = SOUND__DELAY;
							pMemory[address]=val;
							pROM[address&0x7fff]=val;		// never 
							break;


			case	0xff1d: pMemory[0xff1d] = val; ResetScreen(); break;
			case	0xff1f:	pROM[0x7f1f]=val; pMemory[address]=val; break; 
			case	0xff3e:	ROMACTIVE=1; pROM[0x7f3e] = pMemory[address]=val; break; 
			case	0xff3f:	ROMACTIVE=0; pROM[0x7f3f] = pMemory[address]=val; break; 
			default:{
					pMemory[address]=val;
					pROM[address&0x7fff]=val;		// never 
					break;
			}
		}
	}
	else if( address == 0x1 ){
		pMemory[ 0x0001 ] = val | ((val&0x4)<<5);
	}

}

/*//****************************************************************************
// Get a word from memory and move Program counter
//****************************************************************************
int GETAWORD( )
{
	int a = Deek(__PC);
	__PC = (__PC+2)&0xffff;
	return a;
}


//****************************************************************************
// Get a byte from memory and move Program counter
//****************************************************************************
int	GETABYTE ( )
{
	int a = Peek(__PC);
	__PC = (__PC+1)&0xffff;
	return a;
}
*/

//****************************************************************************
// Get a word from memory and move Program counter
//****************************************************************************
int GETAWORD( void )
{
	int	a;

	// IS rom active? if not... read from RAM
	if ((ROMACTIVE==0) || (__PC<0x8000)) {
		a=(pMemory[__PC]&0xff);
		a+=(((pMemory[(__PC+1)&0xffff])&0xff) <<8);
		__PC=(__PC+2)&0xffff;
		return a;
	}else{
		if( ( __PC>=0xfc00 ) && ( __PC<=0xfcff ) ){
			a = (pROM[__PC&0x7fff])|(pROM[(__PC+1)&0x7fff]<<8);
		}
		else if( __PC<0xc000 ){
			a = pRomLo[__PC&0x3fff];
			a |= (pRomLo[(__PC+1)&0x3fff]<<8);
		}
		else{
			a = pRomHi[__PC&0x3fff];
			a |= (pRomHi[(__PC+1)&0x3fff]<<8);
		}
		__PC=(__PC+2)&0xffff;
		return a;
	}
}

//****************************************************************************
// Get a byte from memory and move Program counter
//****************************************************************************
int	GETABYTE ( void )
{
	int		a;
	// IS rom active? if not... read from RAM
	if ((ROMACTIVE==0) || (__PC<0x8000)) {
		a = pMemory[__PC];
		__PC = (__PC+1)&0xffff;
		return (a);
	}else{
//		int a=pROM[__PC-0x8000];
		if( ( __PC>=0xfc00 ) && ( __PC<=0xfcff ) ){
			a = pROM[__PC&0x7fff];
		}
		else if( __PC<0xc000 ){
			a = pRomLo[__PC&0x3fff];
		}
		else{
			a = pRomHi[__PC&0x3fff];
		}
		__PC = (__PC+1)&0xffff;
		return a;
	}
}


//****************************************************************************
// Process the "SEI" instruction
//****************************************************************************
void ProcessSEI( )
{
	__F|=__I;
	Cycles=2;
}

//****************************************************************************
// Process the "HLT" instruction
//****************************************************************************
final void ProcessHLT ( )
{
	__PC=(__PC-1)&0xffff;			// CRASH!!!
	Cycles=2;
}

//****************************************************************************
//
//****************************************************************************
int AddDecimal ( int d1, uint8 a )
{
	int d,c,ans;

	d=d1&0xf;               // get first digit
	d+=(a&0xf);
	if (d>9) {
		ans=d-10;
		c=1;
	}else {
		c=0;				// get half carry
		ans=d;
	}
	d=(d1&0xf0)>>4;			// get first digit
	d+=(a&0xf0)>>4;
	d+=c;					// add half carry
	if (d>9){
		ans+=((d-10)<<4)+0x100;
	} else {
		ans+=(d<<4);
	}
	
	return  ans;
}


//****************************************************************************
//
// Process the "ADC" instruction 
//
// V=Sm & Dm & (~Rm) | (~Sm) & (~Dm) & (Rm)
//
//****************************************************************************
void ProcessADC( )
{
	int		__D1=0;
//	uint16  __D1;
	uint16  ___F;
	int		address,a,c;
	int		zpage;
	int		data;
	int     test;
    
	//___F=__F;
	___F=__F;
	__A&=0xff;
	switch (opcode) {
		case 0x69	:	__D1=(data=GETABYTE()); Cycles=2; break;
		case 0x65   :	zpage=GETABYTE(); __D1=(data=Peek(zpage)); Cycles=3; break;
		case 0x75   :	zpage=GETABYTE(); __D1=(data=Peek((zpage+__X)&255)); Cycles=4; break;
		case 0x6D   :	address=GETAWORD(); __D1=(data=Peek(address)); Cycles=4; break;
		case 0x7D   :	{
							address=GETAWORD(); 
							a=address&0xff00;
							__D1=(data=Peek(address+__X)); 
							if (((address+__X)&0xff00)!=a){
								Cycles=5;                                               //page cross+1
							}else {
								Cycles=4; 
							}
							break; 
						}
		case 0x79	:	{       
							address=GETAWORD(); 
							a=address&0xff00;
							__D1=(data=Peek(address+__Y)); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=5;                                               //page cross+1
							}else {
								Cycles=4; 
							}
							break; 
						}
	   case 0x61    :	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __D1+=(data=Peek(address));    Cycles=6; break; 
	   case 0x71    :	{
							zpage=GETABYTE(); 
							address=Deek(zpage); 
							a=address&0xff00;                                       // need to check for page boundary
							__D1=(data=Peek(address+__Y));
							if (((address+__Y)&0xff00)!=a){
								Cycles=6;                                                       //page cross+1
							}else{
								Cycles=5; 
							}
							break; 
						}
		}
/*		//
		// do actual ADD!
		//
		if ((__F&__D)!=0){                // decimal mode?
				if ((__F&__C)!=0) c=1; else c=0;         
				__D1=AddDecimal(__D1,(uint8) c);  // add carry to decimal number
				if ((__D1&0x100)!=0) c=1; else c=0;
				__D1&=0xff;
				__D1=AddDecimal(__D1,(uint8)__A); // add carry to decimal number
				if (c==1) __D1|=0x100;
		} else {
				__D1+=__A;
				if ((__F&__C)!=0) __D1++;				// if carry set, add one more
		}
		// do flags
		__F&=(__I+__D);								// clear carry as well
		if ((__D1&0x100)!=0) __F|=__C;					// carry set?

		if ((__D1<=127) && (__D1>=-128)) {			// overflow set?
		} else {
			__F|=__V;
		}

		__A=(__D1&255);									// set NEW "Accum"
		if ((__A & 128)==128) {__F|=__N;}
		if ((__A&0xff)==0) {__F|=__Z;}

*/
//******************************
        test=(signed char)__D1+(signed char)__A;
        
        // do actual ADD!
        if ((__F&__D)!=0){                // decimal mode?
                if ((___F&__C)!=0) c=1; else c=0;         
                __D1=AddDecimal(__D1,c);  // add carry to decimal number
                if ((__D1&0x100)!=0) c=1; else c=0;
                __D1&=0xff;
                __D1=AddDecimal(__D1,__A); // add carry to decimal number
                if (c==1) __D1|=0x100;
        } else {
                __D1+=__A;
                if ((___F&__C)!=0) __D1++;
        }

        if ((___F&__C)!=0) test++;


        // do flags
        ___F&=(__I+__D);                                                           // clear carry as well
        if ((__D1&0x100)!=0)___F|=__C;
        if ((test<=127) && (test>=-128)) {
        } else {
            ___F|=__V;
        }
        __A=__D1&255;
        if ((__A & 128)==128) {___F|=__N;}
        if (__A==0) {___F|=__Z;}
        __F=___F;

}


//****************************************************************************
//
// Process the "LSR" instruction    __N=0     0->__A->__C
//
//	ALR    ***
//	This opcode ANDs the contents of the A register with an immediate value and 
//	then LSRs the result.
//
//	One supported mode:
//
//	ALR #ab         ;4B ab       ;No. Cycles= 2
//
//****************************************************************************
void ProcessALR( )
{
	int	data;

	__F&=(__V+__D+__I);

	data=__A&GETABYTE(); 

	if ((data&1)==1) { __F|=__C; }
	data=data>>1;
	if ((data&0xff)==0)  { __F|=__Z; }

	__A=data;
}


//****************************************************************************
// Process the "AND" instruction 
//****************************************************************************
void ProcessAND( )
{
	int		address=0,a=0;
	int		zpage=0;


	switch (opcode) {
	   case 0x29:	__A=__A&GETABYTE(); Cycles=2; break;
	   case 0x25:	zpage=GETABYTE(); __A&=Peek(zpage);        Cycles=3; break;
	   case 0x35:	zpage=GETABYTE(); __A&=Peek((zpage+__X)&255);     Cycles=4; break;
	   case 0x2D:	address=GETAWORD(); __A&=Peek(address);    Cycles=4; break;
	   case 0x3D:	{
						address=GETAWORD(); 
						a=address&0xff00;
						__A&=Peek(address+__X); 
						if (((address+__X)&0xff00)!=a){
							Cycles=5; 							// +1 for crossing page boundary
						} else {
							Cycles=4; 
						}
						break;
					}
	   case 0x39:	{
						address=GETAWORD(); 
						a=address&0xff00;
						__A&=Peek(address+__Y); 
						if (((address+__Y)&0xff00)!=a){
							Cycles=5; 							// +1 for crossing page boundary
						} else {
							Cycles=4; 
						}
						break;
					}
	   case 0x21:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __A&=Peek(address);    Cycles=6; break;
	   case 0x31:	{
						zpage=GETABYTE(); 
						address=Deek(zpage);    
						a=address&0xff00;
						__A&=Peek(address+__Y); 
						if (((address+__Y)&0xff00)!=a){
							Cycles=6; 							// +1 for crossing page boundary
						} else {
							Cycles=5; 
						}
						break;
					}
	}
	COMP (__A);
}


//****************************************************************************
// Process the "AND" instruction 
//****************************************************************************
void ProcessANC( )
{
	int		address=0,a=0;
	int		zpage=0;


	__A=__A&GETABYTE(); 
	Cycles=2;
	COMP (__A);

	__F &= ~__C;
	if( ( __F & __N ) != 0 ){
		__F |= __C;
	}

}

//****************************************************************************
// Process the "ASL" instruction 
//****************************************************************************

void ProcessASL( )
{
	int		address=0;
	int		data=0,zpage=0;

	__F&=(__V+__D+__I);
	switch (opcode) {
	   case 0x0A	:	data=__A; break;
	   case 0x06	:	zpage=GETABYTE(); data=Peek(zpage); 	      break;
	   case 0x16	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
	   case 0x0E	:	address=GETAWORD(); data=Peek(address); Poke(address,data);   break;
	   case 0x1E	:	address=GETAWORD(); data=Peek(address+__X); break;
	}

	if ((data&128)==128) { __F|=__C; }
	data=data<<1;
	data&=0xfe;
	if (data==0)  { __F|=__Z; }
	if ((data&128)==128) { __F|=__N; }

	switch (opcode) {
	   case 0x0A	:	__A=data; Cycles=2; break;
	   case 0x06	:	Poke(zpage,data); 	Cycles=5; break;
	   case 0x16	:	Poke((zpage+__X)&255,data); Cycles=6; break;
	   case 0x0E	:	Poke(address,data);  Cycles=6; break;
	   case 0x1E	:	Poke(address+__X,data); Cycles=7; break;
	}
}

//****************************************************************************
//
// Process the "ASL" instruction 
//
// ASO    ***    (SLO)
// This opcode ASLs the contents of a memory location and then ORs the result 
// with the accumulator.  
// 
// Supported modes:
// 
// ASO abcd        ;0F cd ab    ;No. Cycles= 6
// ASO abcd,X      ;1F cd ab    ;            7
// ASO abcd,Y      ;1B cd ab    ;            7
// ASO ab          ;07 ab       ;            5
// ASO ab,X        ;17 ab       ;            6
// ASO (ab,X)      ;03 ab       ;            8
// ASO (ab),Y      ;13 ab       ;            8
//
//****************************************************************************
void ProcessASO( )
{
	int	address=0;
	int	zpage=0;

	__F&=(__V+__D+__I);
	switch (opcode) {
	   case 0x0F	:	address=GETAWORD(); data=Peek(address);     Cycles=6; break;
	   case 0x1F	:	address=GETAWORD()+__X; data=Peek(address);  Cycles=7; break;
	   case 0x1B	:	address=GETAWORD()+__Y; data=Peek(address);  Cycles=7; break;
	   case 0x07	:	address=GETABYTE(); data=Peek(address); 	  Cycles=5; break;
	   case 0x17	:	address=GETABYTE()+__X; data=Peek(address);  Cycles=6; break;
	   case 0x03	:	address=(GETABYTE()+__X)&255; data=Peek(address); Cycles=8; break;
	   case 0x13	:	zpage=GETABYTE(); address=Deek(zpage)+__Y;    data=Peek(address); Cycles=8; break;
	}

	if ((data&128)==128) { __F|=__C; }
	data=data<<1;
	if ((data&0xff)==0)  { __F|=__Z; }
	if ((data&128)==128) { __F|=__N; }

	Poke(address,data);
	__A=(__A|data)&255;
}

//****************************************************************************
//
// Process the "AXA" instruction 
//
// AXA    ***
// This opcode stores the result of A AND X AND the high byte of the target 
// address of the operand +1 in memory.
//
// Supported modes:
// 
// AXA abcd,Y      ;9F cd ab    ;No. Cycles= 5
// AXA (ab),Y      ;93 ab       ;            6
//
//****************************************************************************
void ProcessAXA( )
{
	int		address=0,a=0;
	int		zpage=0;


	switch (opcode) {
	   case 0x9F:	{
						address=((GETAWORD()+__Y)&0xffff);  
						__A = (__A&__X&(address>>8))&0xff; 
						Cycles = 5; 
						break;
					}
	   case 0x93:	{
						zpage = GETABYTE(); 
						address = ((Deek(zpage)+__Y)&0xffff);    
						__A = (__A&__X&(address>>8))&0xff; 
						Cycles = 5; 
						break;
					}
	}
	COMP (__A);
}

//****************************************************************************
//
// Process the "AXS" instruction 
//
//		AXS abcd        ;8F cd ab    ;No. Cycles= 4
//		AXS ab          ;87 ab       ;            3
//		AXS ab,Y        ;97 ab       ;            4
//		AXS (ab,X)      ;83 ab       ;            6
//
//****************************************************************************
void ProcessAXS( )
{

	int	address=0;
	int	zpage=0;

	switch (opcode) {
	   case 0x87	:	Poke(GETABYTE(),__A&__X); Cycles=3; break;
	   case 0x97	:	Poke((GETABYTE()+__Y)&255,__A&__X);  Cycles=4; break;
	   case 0x8F	:	Poke(GETAWORD(),__A&__X);  Cycles=4; break;
	   case 0x83	:	address=Deek((GETABYTE()+__X)&255); Poke(address,__A&__X); Cycles=6; break;
	}
}


//****************************************************************************
// Process the "BCC" instruction 
//****************************************************************************
final void ProcessBCC( )
{
	int	offset;

	offset=GETABYTE(); 
	if ((__F&__C)==0){
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	} else{
		Cycles=2; 
	}
}

//****************************************************************************
// Process the "BCS" instruction 
//****************************************************************************
void ProcessBCS( )
{
	int  offset;

	offset=GETABYTE(); 
	if ((__F&__C)==__C){
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	}else{
		Cycles=2; 
	}

}

//****************************************************************************
// Process the "BEQ" instruction 
//****************************************************************************
void ProcessBEQ( )
{
	int	offset;

	offset=GETABYTE(); 
	if ((__F&__Z)==__Z){
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	}else{
		Cycles=2; 
	}
}

//****************************************************************************
// Process the "BIT" instruction 
//****************************************************************************
void ProcessBIT( )
{
	int		address=0;
	int		zpage=0,data=0;

	__F&=(__D|__I|__C|__B);
	switch (opcode) {
	   case 0x24	:	zpage=GETABYTE(); data=Peek(zpage); Cycles=3; break;
	   case 0x2c	:	address=GETAWORD(); data=Peek(address); Cycles=4; break;
	}
	
	//__A&=0xff;
	if ((data&__A)==0) __F|=__Z;
	if ((data&128)!=0) __F|=__N;
	if ((data&64)!=0) __F|=__V;
}

//****************************************************************************
//
// Process the "BMI" instruction 
//
//****************************************************************************
void ProcessBMI( )
{
	int	offset;

	offset=GETABYTE(); 
	if ((__F&__N)==__N) {
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	}else{
		Cycles=2; 
	}
}

//****************************************************************************
// Process the "BNE" instruction 
//****************************************************************************
void ProcessBNE( )
{
	int	offset;

	offset=GETABYTE();
	if ((__F&__Z)==0) {
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	}else{
		Cycles=2; 
	}
}

//****************************************************************************
// Process the "BPL" instruction 
//****************************************************************************
void  ProcessBPL( )
{
	int	offset;

	offset=GETABYTE(); 
	if ((__F&__N)==0) {
		int	NewPC;
		
		NewPC=__PC&0xff00;
		TakeBranch(offset);
		if ((__PC&0xff00)!=NewPC){
			Cycles=4;					// Branch taken + Page boundary crossing
		} else {
			Cycles=3;					// just +1 for branch taken
		}
	}else{
		Cycles=2; 
	}
}

//****************************************************************************
//
// Process the "BRK" instruction 
//
//****************************************************************************
void ProcessBRK( )
{
	int	var;

	var=GETABYTE();                         // ignore just now.
	Cycles=7;
	DushReg (__PC);
	PushReg (__F);
	__F|=__I|__B;
	__PC=Deek(0xfffe);                       //0xfcb3;
}

//****************************************************************************
// Process the "BVC" instruction 
//****************************************************************************
	void ProcessBVC( )
	{
		int	offset;

		offset=GETABYTE(); 
		if ((__F&__V)==0) {
			int	NewPC;
			
			NewPC=__PC&0xff00;
			TakeBranch(offset);
			if ((__PC&0xff00)!=NewPC){
				Cycles=4;					// Branch taken + Page boundary crossing
			} else {
				Cycles=3;					// just +1 for branch taken
			}
		}else{
			Cycles=2; 
		}
	}

	//****************************************************************************
	//
	// Process the "BVS" instruction 
	//
	//****************************************************************************
	void ProcessBVS( )
	{
		int	offset;

		offset=GETABYTE(); 
		if ((__F&__V)==__V) {
			int	NewPC;
			
			NewPC=__PC&0xff00;
			TakeBranch(offset);
			if ((__PC&0xff00)!=NewPC){
				Cycles=4;					// Branch taken + Page boundary crossing
			} else {
				Cycles=3;					// just +1 for branch taken
			}
		}else{
			Cycles=2; 
		}
	}

	//****************************************************************************
	// Process the "CLC" instruction 
	//****************************************************************************
	void ProcessCLC( )
	{
		__F&=(~__C);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "CLD" instruction 
	//****************************************************************************
	void ProcessCLD( )
	{
		__F&=(~__D);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "CLI" instruction 
	//****************************************************************************
	void ProcessCLI( )
	{
		__F&=(~__I);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "CLV" instruction 
	//****************************************************************************
	void ProcessCLV( )
	{
		__F&=(~__V);
		Cycles=2; 
	}

//****************************************************************************
//
// Process the "CMP" instruction 
//
//****************************************************************************
void ProcessCMP( )
{
	int	address,a;
	int	zpage;
	int	test;

	switch (opcode) {
	   case 0xC9:	data=GETABYTE(); Cycles=2; break;
	   case 0xC5:	zpage=GETABYTE(); data=Peek(zpage); Cycles=3; break;
	   case 0xD5:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); Cycles=4; break;
	   case 0xCD:	address=GETAWORD(); data=Peek(address); Cycles=4; break;
	   case 0xDD:	{
						address=GETAWORD();
						a=address&0xff00;
						data=Peek(address+__X);
						if (((address+__X)&0xff00)!=a){
							Cycles=5; 							// +1 for crossing page boundary
						} else {
							Cycles=4;
						}
						break;
					}
	   case 0xD9:	{
						address=GETAWORD(); 
						a=address&0xff00;
						data=Peek(address+__Y); 
						if (((address+__Y)&0xff00)!=a){
							Cycles=5; 							// +1 for crossing page boundary
						} else {
							Cycles=4; 
						}
						break;
					}
	   case 0xC1:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address);    Cycles=6; break;
	   case 0xD1:	{
						zpage=GETABYTE(); 
						address=Deek(zpage);
						a=address&0xff00;
						data=Peek(address+__Y); 
						if (((address+__Y)&0xff00)!=a){
							Cycles=6; 							// +1 for crossing page boundary
						} else {
							Cycles=5; 
						}
						break;
					}
	}

	__A&=0xff;
	__F&=(__I+__D+__V+__B);					// wipe flags
	if (__A>=data) {__F|=__C;}				// carry set when a>=data (clear = a<data)
	test=__A-data;
	if ((test&128)!=0) {__F|=__N;}		// sign bit set?
	if ((test&0xff)==0) {__F|=__Z;}

}

	//****************************************************************************
	// Process the "CPX" instruction 
	//****************************************************************************
	void ProcessCPX( )
	{
		int		data,test;
		
		switch (opcode) {
			case 0xe0:	data=GETABYTE();		Cycles=2; break;
			case 0xe4:	data=Peek(GETABYTE());	Cycles=3; break;
			case 0xec:	data=Peek(GETAWORD());	Cycles=4; break;
			default: data=0;
		}

		__X&=0xff;
		__F&=(__I+__D+__V+__B);					// wipe flags
		if (__X>=data) {__F|=__C;}				// carry set when a>=data (clear = a<data)
		test=__X-data;
		if ((test&128)!=0) {__F|=__N;}		// sign bit set?
		if ((test&0xff)==0) {__F|=__Z;}
	}

	//****************************************************************************
	// Process the "CPY" instruction 
	//****************************************************************************
	void ProcessCPY( )
	{
		int		address;
		int		zpage;
		int		data=0;
		int		test;
		
		switch (opcode) {
		   case 0xC0	:	data=GETABYTE(); Cycles=2; break;
		   case 0xC4	:	zpage=GETABYTE(); data=Peek(zpage);        Cycles=3; break;
		   case 0xCC	:	address=GETAWORD(); data=Peek(address);    Cycles=4; break;
		}

		__Y&=0xff;
		__F&=(__I+__D+__V+__B);					// wipe flags
		if (__Y>=data) { __F|=__C; }			// carry set when a>=data (clear = a<data)
		test = (__Y-data);
		if ((test&128)!=0) {__F|=__N;}		// sign bit set?
		if ((test&0xff)==0) {__F|=__Z;}
	}

	//****************************************************************************
	//
	// Process the "DCM" instruction 
	//
	//	DCM    ***    (DCP)
	//	This opcode DECs the contents of a memory location and then CMPs the result 
	//	with the A register.
	//
	//	Supported modes:
	//
	//	DCM ab          ;C7 ab       ;            5
	//	DCM ab,X        ;D7 ab       ;            6
	//	DCM abcd        ;CF cd ab    ;No. Cycles= 6
	//	DCM abcd,X      ;DF cd ab    ;            7
	//	DCM abcd,Y      ;DB cd ab    ;            7
	//	DCM (ab,X)      ;C3 ab       ;            8
	//	DCM (ab),Y      ;D3 ab       ;            8
	//
	//****************************************************************************
	void ProcessDCM( )
	{
		int		address;
		int		zpage;
		int		test;

		switch (opcode) {
		   case 0xC7	:	zpage=GETABYTE(); data=Peek(zpage)-1; Poke(zpage,data);       Cycles=5; break;
		   case 0xD7	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255)-1;  Poke((zpage+__X)&255,data);   Cycles=6; break;
		   case 0xCF	:	address=GETAWORD(); data=Peek(address)-1; Poke(address,data);    Cycles=6; break;
		   case 0xDF	:	address=GETAWORD(); data=Peek(address+__X)-1; Poke (address+__X,data); Cycles=7; break;
		   case 0xDB	:	address=GETAWORD(); data=Peek(address+__Y)-1; Poke (address+__Y,data);Cycles=7; break;
		   case 0xC3	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address)-1; Poke(address,data);   Cycles=8; break;
		   case 0xD3	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+__Y)-1; Poke (address+__Y,data); Cycles=8; break;
		}

		__F&=(__I+__D+__V+__B);					// wipe flags
		if (__A>=data) {__F|=__C;}				// carry set when a>=data (clear = a<data)
		test=__A-data;
		if ((test&128)!=0) {__F|=__N;}		// sign bit set?
		if ((test&0xff)==0) {__F|=__Z;}
	}

	//****************************************************************************
	// Process the "DEC" instruction 
	//****************************************************************************
	void ProcessDEC( )
	{
		int		address;
		int		zpage;
		int		data=0;

		switch (opcode) {
		   case 0xC6:	zpage=GETABYTE(); data=Peek(zpage)-1; Poke(zpage,data); Cycles=5; break;
		   case 0xD6:	address=(GETABYTE()+__X)&0xff; data=Peek(address)-1; Poke(address,data); Cycles=6; break;
		   case 0xCE:	address=GETAWORD(); data=Peek(address); Poke(address,data); data--;	  Poke(address,data); Cycles=6; break;
		   case 0xDE:	address=GETAWORD(); data=Peek(address+__X)-1; Poke(address+__X,data); Cycles=7; break;
		}
		data&=0xff;
		COMP(data);
	}

	//****************************************************************************
	// Process the "DEX" instruction 
	//****************************************************************************
	final void ProcessDEX( )
	{
		__X--;
		__X&=0xff;
		COMP(__X);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "DEY" instruction 
	//****************************************************************************
	final void ProcessDEY( )
	{
		__Y--;
		__Y&=0xff;
		COMP(__Y);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "EOR" instruction 
	//****************************************************************************
	void ProcessEOR( )
	{
		int		address,a;
		int		zpage;

		switch (opcode) {
		   case 0x49:	__A^=GETABYTE(); Cycles=2; break;
		   case 0x45:	zpage=GETABYTE(); __A^=Peek(zpage);        Cycles=3; break;
		   case 0x55:	zpage=GETABYTE(); __A^=Peek((zpage+__X)&255);     Cycles=4; break;
		   case 0x4D:	{
							address=GETAWORD(); 
							__A^=Peek(address);    Cycles=4; break;
						}
		   case 0x5D:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A^=Peek(address+__X); 
							if (((address+__X)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0x59:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A^=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0x41:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __A^=Peek(address);    Cycles=6; break;
		   case 0x51:	{
							zpage=GETABYTE(); 
							address=Deek(zpage);    
							a=address&0xff00;
							__A^=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=6; 							// +1 for crossing page boundary
							} else {
								Cycles=5; 
							}
							break;
						}
		}
		COMP(__A);
	}

	//****************************************************************************
	// Process the "INC" instruction 
	//****************************************************************************
	void ProcessINC( )
	{
		int		address;
		int		zpage;

		switch (opcode) {
		   case 0xE6	:	zpage=GETABYTE(); Poke(zpage,(data=Peek(zpage)+1));       Cycles=5; break;
		   case 0xF6	:	zpage=GETABYTE(); Poke((zpage+__X)&255,(data=Peek((zpage+__X)&255)+1)); Cycles=6; break;
		   case 0xEE	:	address=GETAWORD(); 
							   data=Peek(address);
								Poke (address,data);			// allow "inc $ff09"
								data++;
								Poke(address,data); 
								Cycles=6; 
								break;
		   case 0xFE	:	address=GETAWORD()+__X; Poke(address,(data=Peek(address)+1)); Cycles=7; break;
		}
		COMP(data);
	}

	//****************************************************************************
	// Process the "SBC" instruction 
	//****************************************************************************
	void ProcessINS( )
	{
		int		address;
		int		zpage,t,last;


		t=1;
		if ((__F&__C)!=0) t--;
		switch (opcode) {
		   case 0xe7	:	zpage=GETABYTE(); data=Peek(zpage)+1; Poke(zpage,data); Cycles=5; break;
		   case 0xf7	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255)+1; Poke((zpage+__X)&255,data);     Cycles=6; break;
		   case 0xef	:	address=GETAWORD(); data=Peek(address)+1;    Poke(address,data);    Cycles=6; break;
		   case 0xff	:	address=GETAWORD(); data=Peek(address+__X)+1; Poke(address+__X,data); Cycles=7; break;
		   case 0xfb	:	address=GETAWORD(); data=Peek(address+__Y)+1; Poke(address+__Y,data); Cycles=7; break;
		   case 0xe3	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address)+1; Poke(address,data); Cycles=8; break;
		   case 0xf3	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+__Y)+1; Poke(address,data); Cycles=8; break;
		}
		__F&=(__D+__I+__C);
		__F|=__C;
		last=(__A-data-t);
		if (last>__A) __F&=(~__C);				// Clear Carry
		if (last==__A) {
			if (((data&0xff)==0xff) && (t==1))  __F&=(~__C);
		}
		__A=last;
		COMP(__A);
	}

	//****************************************************************************
	// Process the "INX" instruction 
	//****************************************************************************
	final void ProcessINX( )
	{
		__X++;
		__X&=0xff;
		COMP(__X);
		Cycles=2; 
	}

	//****************************************************************************
	//
	// Process the "INY" instruction 
	//
	//****************************************************************************
	final void ProcessINY( )
	{
		__Y=(__Y+1)&0xff;
		COMP(__Y);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "JMP" instruction 
	//****************************************************************************
	final void ProcessJMP( )
	{
		int	address;
		

		switch (opcode) {
		   case 0x4C	:	address=GETAWORD(); __PC=address;		Cycles=3; break;
		   case 0x6C	:	address=GETAWORD(); __PC=Deek(address);	Cycles=5; break;
		}
	}


	//****************************************************************************
	// Process the "JSR" instruction 
	//****************************************************************************
	final void ProcessJSR( )
	{
		int	address;
		
		address=GETAWORD();
		DushReg((__PC-1));
		__PC=address;
		Cycles=6; 
	}

	//****************************************************************************
	//
	// Process the "LDA" instruction 
	//
	//	LAX ab          ;A7 ab       ;*=add 1     3
	//	LAX ab,X        ;B7 ab       ;if page     4
	//	LAX abcd        ;AF cd ab    ;No. Cycles= 4
	//	LAX abcd,Y      ;BF cd ab    ;            4*
	//	LAX (ab,X)      ;A3 ab       ;boundary    6
	//	LAX (ab),Y      ;B3 ab       ;is crossed  5*
	//
	//****************************************************************************
	void ProcessLAX( )
	{
		int		address;
		int		zpage;

		switch (opcode) {
		   case 0xA7	:	zpage=GETABYTE(); __A=Peek(zpage); __X=__A; Cycles=3; break;
		   case 0xB7	:	zpage=GETABYTE(); __A=Peek((zpage+__X)&255); __X=__A;   Cycles=4; break;
		   case 0xAF	:	address=GETAWORD(); __A=Peek(address);  __X=__A; Cycles=4; break;
		   case 0xBF	:	address=GETAWORD(); __A=Peek(address+__Y); __X=__A; Cycles=4; break;
		   case 0xA3	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __A=Peek(address); __X=__A; Cycles=6; break;
		   case 0xB3	:	zpage=GETABYTE(); address=Deek(zpage);    __A=Peek(address+__Y); __X=__A; Cycles=5; break;
		}
		COMP(__A);
	}


	//****************************************************************************
	// Process the "LDA" instruction 
	//****************************************************************************
	final void ProcessLDA( )
	{
		int		address,a;
		int		zpage;

		switch (opcode) {
//		   case 0xA9:	__A=GETABYTE(); Cycles=2; break;
//		   case 0xA5:	zpage=GETABYTE(); __A=Peek(zpage);        Cycles=3; break;
//		   case 0xB5:	zpage=GETABYTE(); __A=Peek((zpage+__X)&255);     Cycles=4; break;
//		   case 0xAD:	address=GETAWORD(); __A=Peek(address);    Cycles=4; break;

		   case 0xBD:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A=Peek(address+__X); 
							if (((address+__X)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0xB9:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0xA1:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __A=Peek(address);    Cycles=6; break;
		   case 0xB1:	{
							zpage=GETABYTE(); 
							address=Deek(zpage);    
							a=address&0xff00;
							__A=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=6; 							// +1 for crossing page boundary
							} else {
								Cycles=5; 
							}
							break;
						}
		}

		COMP(__A);
	}

	//****************************************************************************
	// Process the "LDX" instruction 
	//****************************************************************************
	final void ProcessLDX( )
	{
		int		address,a;
		//int		zpage;

		//switch (opcode) {
		   //case 0xA2:	__X=GETABYTE(); Cycles=2; break;
		   //case 0xA6:	zpage=GETABYTE(); __X=Peek(zpage);        Cycles=3; break;
		   //case 0xB6:	zpage=GETABYTE(); __X=Peek((zpage+__Y)&0xff);     Cycles=4; break;
		   //case 0xAE:	address=GETAWORD(); __X=Peek(address);    Cycles=4; break;
		   //case 0xBE:	
			   {
							address=GETAWORD(); 
							a=address&0xff00;
							__X=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							//break;
						}
		//}
		COMP(__X);
	}


	//****************************************************************************
	// Process the "LDY" instruction 
	//****************************************************************************
	final void ProcessLDY( )
	{
		int		address,a;
//		int		zpage;

//		switch (opcode) {
		   //case 0xA0:	__Y=GETABYTE(); Cycles=2; break;
		   //case 0xA4:	zpage=GETABYTE(); __Y=Peek(zpage);        Cycles=3; break;
		   //case 0xB4:	zpage=GETABYTE(); __Y=Peek((zpage+__X)&255);     Cycles=4; break;
		   //case 0xAC:	address=GETAWORD(); __Y=Peek(address);    Cycles=4; break;
//		   case 0xBC:	
		{
							address=GETAWORD(); 
							a=address&0xff00;
							__Y=Peek(address+__X); 
							if (((address+__X)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
//							break;
						}
//		}																			
		COMP(__Y);
	}


	//****************************************************************************
	//
	// Process the "LSE" instruction    __N=0     0->__A->__C
	//
	//	LSE    ***   (SRE)
	//	LSE LSRs the contents of a memory location and then EORs the result with 
	//	the accumulator.
	//
	//	Supported modes:
	//
	//	LSE ab          ;47 ab       ;            5
	//	LSE ab,X        ;57 ab       ;            6
	//	LSE abcd        ;4F cd ab    ;No. Cycles= 6
	//	LSE abcd,X      ;5F cd ab    ;            7
	//	LSE abcd,Y      ;5B cd ab    ;            7
	//	LSE (ab,X)      ;43 ab       ;            8
	//	LSE (ab),Y      ;53 ab       ;            8
	//
	//****************************************************************************
	void ProcessLSE( )
	{
		int		address=0;
		int		zpage=0;

		__F&=(__V+__D+__I);
		switch (opcode) {
		   case 0x47	:	zpage=GETABYTE(); data=Peek(zpage); 	  break;
		   case 0x57	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x4F	:	address=GETAWORD(); data=Peek(address); break;
		   case 0x5F	:	address=GETAWORD(); data=Peek(address+__X); break;
		   case 0x5B	:	address=GETAWORD(); data=Peek(address+__Y); break;
		   case 0x43	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address);    Cycles=6; break;
		   case 0x53	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+__Y); Cycles=5; break;
		}

		if ((data&1)==1) { __F|=__C; }
		data=data>>1;
		if ((data&0xff)==0)  { __F|=__Z; }

		switch (opcode) {
		   case 0x47	:	Poke(zpage,data);    Cycles=5; break;
		   case 0x57	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x4F	:	Poke(address,data);  Cycles=6; break;
		   case 0x5F	:	Poke(address+__X,data); Cycles=7; break;
		   case 0x5B	:	Poke(address+__Y,data); Cycles=7; break;
		   case 0x43	:	Poke(address,data);    Cycles=8; break;
		   case 0x53	:	Poke(address+__Y,data); Cycles=8; break;
		}
		__A^=data;
	}

	//****************************************************************************
	// Process the "LSR" instruction    __N=0     0->__A->__C
	//****************************************************************************
	void ProcessLSR( )
	{
		int		address=0;
		int		zpage=0,data=0;

		__F&=(__V+__D+__I);
		switch (opcode) {
		   case 0x4A	:	data=__A; break;
		   case 0x46	:	zpage=GETABYTE(); data=Peek(zpage); 	  break;
		   case 0x56	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x4E	:	address=GETAWORD(); data=Peek(address); Poke(address,data); break;
		   case 0x5E	:	address=GETAWORD(); data=Peek(address+__X); break;
		}

		if ((data&1)==1) { __F|=__C; }
		data=data>>1;
		data&=0x7f;
		if ((data&0xff)==0)  { __F|=__Z; }

		switch (opcode) {
		   case 0x4A	:	__A=data; Cycles=2; break;
		   case 0x46	:	Poke(zpage,data);    Cycles=5; break;
		   case 0x56	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x4E	:	Poke(address,data);  Cycles=6; break;
		   case 0x5E	:	Poke(address+__X,data); Cycles=7; break;
		}
	}


	//****************************************************************************
	//
	// Process the "NOP" instruction 
	//
	//****************************************************************************
	final void ProcessNOP( )
	{
		Cycles=2;
	}


	//****************************************************************************
	//
	// Process the "ORA" instruction 
	//
	// OAL #ab         ;AB ab       ;No. Cycles= 2
	//
	//****************************************************************************
	void ProcessOAL( )
	{
		__A|=0xee;
		__A&=GETABYTE(); 
		__X=__A;
		Cycles=2;
		COMP(__A);
	}


	//****************************************************************************
	// Process the "ORA" instruction 
	//****************************************************************************
	void ProcessORA( )
	{
		int		address=0,a;
		int		zpage;


		switch (opcode) {
		   //case 0x09:	__A|=GETABYTE(); Cycles=2; break;
		   //case 0x05:	zpage=GETABYTE(); __A|=Peek(zpage);        Cycles=3; break;
		   case 0x15:	zpage=GETABYTE(); __A|=Peek((zpage+__X)&255);     Cycles=4; break;
		   case 0x0D:	address=GETAWORD(); __A|=Peek(address);    Cycles=4; break;
		   case 0x1D:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A|=Peek(address+__X); 
							if (((address+__X)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0x19:	{
							address=GETAWORD(); 
							a=address&0xff00;
							__A|=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=5; 							// +1 for crossing page boundary
							} else {
								Cycles=4; 
							}
							break;
						}
		   case 0x01:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); __A|=Peek(address);    Cycles=6; break;
		   case 0x11:	{
							zpage=GETABYTE(); 
							a=address&0xff00;
							address=Deek(zpage);    
							__A|=Peek(address+__Y); 
							if (((address+__Y)&0xff00)!=a){
								Cycles=6; 							// +1 for crossing page boundary
							} else {
								Cycles=5; 
							}
							break;
						}
		}
		COMP(__A);
	}



	//****************************************************************************
	// Process the "PHA" instruction 
	//****************************************************************************
	final void ProcessPHA( )
	{
		PushReg (__A);
		Cycles=3; 
	}

	//****************************************************************************
	// Process the "PHP" instruction 
	//****************************************************************************
	final void ProcessPHP( )
	{
		PushReg (__F);
		Cycles=3; 
	}

	//****************************************************************************
	// Process the "PLA" instruction 
	//****************************************************************************
	final void ProcessPLA( )
	{
		__A=Pull ();
		COMP(__A);
		Cycles=4; 
	}

	//****************************************************************************
	// Process the "PHP" instruction 
	//****************************************************************************
	final void ProcessPLP( )
	{
		__F=Pull ();
		Cycles=4; 
	}


	//****************************************************************************
	//
	// Process the "ROL" instruction 
	//
	//	RLA    ***
	//	RLA ROLs the contents of a memory location and then ANDs the result with 
	//	the accumulator.
	//
	//	Supported modes:
	//
	//	RLA ab          ;27 ab       ;            5
	//	RLA ab,X        ;37 ab       ;            6
	//	RLA abcd        ;2F cd ab    ;No. Cycles= 6
	//	RLA abcd,X      ;3F cd ab    ;            7
	//	RLA abcd,Y      ;3B cd ab    ;            7
	//	RLA (ab,X)      ;23 ab       ;            8
	//	RLA (ab),Y      ;33 ab       ;            8
	//
	//****************************************************************************
	void ProcessRLA( )
	{
		int		address=0;
		int		zpage=0;
		int		t;

		t=0;
		if ((__F&__C)!=0) t=1;
		__F&=(__V+__D+__I);
		switch (opcode) {
		   case 0x27	:	zpage=GETABYTE(); data=Peek(zpage);  break;
		   case 0x37	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x2F	:	address=GETAWORD(); data=Peek(address); Poke(address,data);break;
		   case 0x3F	:	address=GETAWORD(); data=Peek(address+__X); break;
		   case 0x3B	:	address=GETAWORD(); data=Peek(address+__Y); break;
		   case 0x23	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address); break;
		   case 0x33	:	zpage=GETABYTE();	address=Deek(zpage); data=Peek(address+__Y); break;
		}

		
		if ((data&128)==128) { __F|=__C; }
		data=(data<<1)|t;
		if ((data&0xff)==0)  { __F|=__Z; }
		if ((data&128)==128) { __F|=__N; }
		__A=__A&data;

		switch (opcode) {
		   case 0x26	:	Poke(zpage,data);  	Cycles=5; break;
		   case 0x36	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x2F	:	Poke(address,data); 	Cycles=6; break;
		   case 0x3F	:	Poke(address+__X,data); Cycles=7; break;
		   case 0x3B	:	Poke(address+__Y,data); Cycles=7; break;
		   case 0x23	:	Poke(address,data); Cycles=8; break;
		   case 0x33	:	Poke(address+__Y,data); Cycles=8; break;
		}
	}


	//****************************************************************************
	//
	// Process the "ROL" instruction 
	//
	//****************************************************************************
	void ProcessROL( )
	{
		int	address=0;
		int	zpage=0;
		int	t;

		t=0;
		if ((__F&__C)!=0) t=1;
		__F&=(__V+__D+__I+__B);
		switch (opcode) {
		   case 0x2A	:	data=__A; break;
		   case 0x26	:	zpage=GETABYTE(); data=Peek(zpage);  break;
		   case 0x36	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x2E	:	address=GETAWORD(); data=Peek(address); Poke(address,data);break;
		   case 0x3E	:	address=GETAWORD(); data=Peek(address+__X); break;
		}

		if ((data&128)==128) { __F|=__C; }
		data=(data<<1);	//|t;
		data&=0xfe;
		data |= t;
		if ((data&0xff)==0)  { __F|=__Z; }
		if ((data&128)==128) { __F|=__N; }

		switch (opcode) {
		   case 0x2A	:	__A=data; Cycles=2; break;
		   case 0x26	:	Poke(zpage,data);  	Cycles=5; break;
		   case 0x36	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x2E	:	Poke(address,data); 	Cycles=6; break;
		   case 0x3E	:	Poke(address+__X,data); Cycles=7; break;
		}
	}


	//****************************************************************************
	//
	// Process the "ROR" instruction 
	//
	//****************************************************************************
	void ProcessROR( )
	{
		int		address=0;
		int		zpage=0;
		int		t;

		t=0;
		if ((__F&__C)!=0) t=128;
		__F&=(__V+__D+__I+__B);
		switch (opcode) {
		   case 0x6A	:	data=__A; break;
		   case 0x66	:	zpage=GETABYTE(); data=Peek(zpage); 	break;
		   case 0x76	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x6E	:	address=GETAWORD(); data=Peek(address); Poke(address,data); break;
		   case 0x7E	:	address=GETAWORD(); data=Peek(address+__X); break;
		}

		
		if ((data&1)==1) { __F|=__C; }
		data=(data>>1);	//|t;
		data &= 0x7f;
		data |= t;
		if ((data&0xff)==0)  { __F|=__Z; }
		if ((data&128)==128) { __F|=__N; }

		switch (opcode) {
		   case 0x6A	:	__A=data; Cycles=2; break;
		   case 0x66	:	Poke(zpage,data);   	Cycles=5; break;
		   case 0x76	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x6E	:	Poke(address,data); 	Cycles=6; break;
		   case 0x7E	:	Poke(address+__X,data); Cycles=7; break;
		}
	}


	//****************************************************************************
	//
	// Process the "RRA" instruction 
	//
	//	RRA    ***
	//	RRA RORs the contents of a memory location and then ADCs the result with 
	//	the accumulator.
	//
	//	Supported modes:
	//
	//	RRA ab          ;67 ab       ;            5
	//	RRA ab,X        ;77 ab       ;            6
	//	RRA abcd        ;6F cd ab    ;No. Cycles= 6
	//	RRA abcd,X      ;7F cd ab    ;            7
	//	RRA abcd,Y      ;7B cd ab    ;            7
	//	RRA (ab,X)      ;63 ab       ;            8
	//	RRA (ab),Y      ;73 ab       ;            8
	//
	//****************************************************************************
	void ProcessRRA( )
	{
		int		address=0;
		int		zpage=0;
		int		t;
		int		d1;

		t=0;
		if ((__F&__C)!=0) t=128;
		__F&=(__V+__D+__I);
		switch (opcode) {
		   case 0x67	:	zpage=GETABYTE(); data=Peek(zpage); 	break;
		   case 0x77	:	zpage=GETABYTE(); data=Peek((zpage+__X)&255); break;
		   case 0x6F	:	address=GETAWORD(); data=Peek(address); break;
		   case 0x7F	:	address=GETAWORD(); data=Peek(address+__X); break;
		   case 0x7B	:	address=GETAWORD(); data=Peek(address+__Y); break;
		   case 0x63	:	zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address);  break;
		   case 0x73	:	zpage=GETABYTE(); address=Deek(zpage);    data=Peek(address+__Y); break;
		}

		
		if ((data&1)==1) { __F|=__C; }
		data=(data>>1)|t;
		if ((data&0xff)==0)  { __F|=__Z; }
		if ((data&128)==128) { __F|=__N; }

		switch (opcode) {
		   case 0x67	:	Poke(zpage,data); Cycles=5; break;
		   case 0x77	:	Poke((zpage+__X)&255,data); Cycles=6; break;
		   case 0x6F	:	Poke(address,data); 	Cycles=6; break;
		   case 0x7F	:	Poke(address+__X,data); Cycles=7; break;
		   case 0x7B	:	Poke(address+__Y,data); Cycles=7; break;
		   case 0x63	:	Poke(address,data);  Cycles=8; break;
		   case 0x73	:	Poke(address+__Y,data); Cycles=8; break;
		}


		//
		// Do ADC part
		//
		d1=__A;
		if ((__F&__C)!=0) d1++;
		d1+=data;								// add data with __A

		// do flags
		__F&=(__I+__D);								// clear carry as well
		if ((d1&0x100)!=0) __F|=__C;
		__A=d1&255;
		if ((__A & 128)==128) {__F|=__N;}
		if ((__A&0xff)==0) {__F|=__Z;}

	}

	//****************************************************************************
	// Process the "RTI" instruction 
	//****************************************************************************
	final void ProcessRTI( )
	{
		__F=Pull ();
		__PC=Dull ();
		Cycles=6;
	}

	//****************************************************************************
	// Process the "RTS" instruction 
	//****************************************************************************
	final void ProcessRTS( )
	{
		__PC=Dull();
		__PC=(__PC+1)&0xffff;
		Cycles=6; 
	}

	//****************************************************************************
	//
	// Process the "CMP" instruction 
	//
	//	SAX    ***
	//	SAX ANDs the contents of the A and X registers (leaving the contents of A 
	//	intact), subtracts an immediate value, and then stores the result in X.
	//	... A few points might be made about the action of subtracting an immediate 
	//	value.  It actually works just like the CMP instruction, except that CMP 
	//	does not store the result of the subtraction it performs in any register.  
	//	This subtract operation is not affected by the state of the Carry flag, 
	//	though it does affect the Carry flag.  It does not affect the Overflow 
	//	flag.
	//
	//	One supported mode:
	//
	//	SAX #ab         ;CB ab       ;No. Cycles= 2
	//
	//	Example:
	//
	//	SAX #$5A        ;CB 5A
	//
	//	Equivalent instructions:
	//
	//	STA $02
	//	TXA
	//	AND $02
	//	SEC
	//	SBC #$5A
	//	TAX
	//	LDA $02
	//
	//****************************************************************************
	void ProcessSAX( )
	{
		int		test,a;

		a=__A&__X;

		data=GETABYTE(); 
		Cycles=2; 

		__F&=(__I+__D+__V+__B);					// wipe flags
		if (a>=data) {__F|=__C;}				// carry set when a>=data (clear = a<data)
		test=a-data;
		__X=test;
		if ((test&128)!=0) {__F|=__N;}		// sign bit set?
		if ((test&0xff)==0) {__F|=__Z;}
	}


	//****************************************************************************
	//
	// Process the "STY" instruction 
	//
	// SAY    ***
	// This opcode ANDs the contents of the Y register with <ab+1> and stores the 
	// result in memory.
	//
	// One supported mode:
	//
	// SAY abcd,X      ;9C cd ab    ;No. Cycles= 5
	//
	// Example:
	//
	// SAY $7700,X     ;9C 00 77
	//
	// Equivalent instructions:
	//
	// PHA
	// TYA
	// AND #$78
	// STA $7700,X
	// PLA
	//
	//****************************************************************************
	void ProcessSAY( )
	{
		int		address;
		int		temp;

		Cycles=5; 
		address=GETAWORD();
		temp=__Y&(((address>>8)+1)&0xff);
		Poke (address+__X,temp);
	}

	//****************************************************************************
	// Process the "SBC" instruction 
	//****************************************************************************
	void ProcessSBC( )
	{      
        int		address,a;
        uint8  	zpage,t,last;
        int 	test;


        t=1;
        if ((__F&__C)!=0) t--;
        switch (opcode) {
           case 0xeb       :       // NOT a normal SBC opcode!!!
           case 0xe9    :       data=GETABYTE(); Cycles=2; break;
           case 0xe5    :       zpage=GETABYTE(); data=Peek(zpage);        Cycles=3; break;
           case 0xf5    :       zpage=GETABYTE(); data=Peek((zpage+__X)&255);     Cycles=4; break;
           case 0xeD    :       address=GETAWORD(); data=Peek(address);    Cycles=4; break;
           case 0xfD    :       {
                                                                address=GETAWORD(); 
                                                                a=address&0xff00;
                                                                data=Peek(address+__X); 
                                                                if (((address+__X)&0xff00)!=a){
                                                                        Cycles=5;                                                       // +1 for crossing page boundary
                                                                } else {
                                                                        Cycles=4; 
                                                                }
                                                                break;
                                                        }
           case 0xf9    :       {
                                                                address=GETAWORD(); 
                                                                a=address&0xff00;
                                                                data=Peek(address+__Y); 
                                                                if (((address+__Y)&0xff00)!=a){
                                                                        Cycles=5;                                                       // +1 for crossing page boundary
                                                                } else {
                                                                        Cycles=4; 
                                                                }
                                                                break;
                                                        }
           case 0xe1    :       zpage=GETABYTE(); address=Deek((zpage+__X)&255); data=Peek(address);    Cycles=6; break;
           case 0xf1    :       {
                                                                zpage=GETABYTE(); 
                                                                address=Deek(zpage);    
                                                                a=address&0xff00;
                                                                data=Peek(address+__Y); 
                                                                if (((address+__Y)&0xff00)!=a){
                                                                        Cycles=6;                                                       // +1 for crossing page boundary
                                                                } else {
                                                                        Cycles=5; 
                                                                }
                                                                break;
                                                        }
        }
        __F&=(__D+__I+__C);
        __F|=__C;


        test=(signed char)__A-(signed char)data-(signed char)t;
        test=test+128;
        
        last=(__A-data-t);
//        if (((__A&0x80)==0x80) && ((last&0x80)==0)) __F|=__V;
        if (test==(test&255)) {
//        if ((test<=127) && (test>=-128)) {
        } else {
            __F|=__V;
        }
        if (last>__A) __F&=(~__C);                         // Clear Carry
        if (last==__A) {
                if ((data==0xff) AND (t==1)) __F&=(~__C);
        }

        //  Decimal mode SBC
        if (__F&__D) {
            unsigned int decres;

            decres=(__A&0xf)-(data&0xf)-t;
            if (decres&0x10) {
                decres=((decres-6)&0xf) | ((__A&0xf0)-(data&0xf0)-0x10);
            } else {
                decres=(decres&0xf) | ((__A&0xf0)-(data&0xf0));
            }
            if (decres&0x100) {
                decres-=0x60;
            }
            __A=decres&0xff;
        } else {
            __A=last;
        }
        COMP(__A);
        

	}

	//****************************************************************************
	// Process the "SEC" instruction 
	//****************************************************************************
	final void ProcessSEC( )
	{
		__F|=__C;
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "SED" instruction 
	//****************************************************************************
	final void ProcessSED( )
	{
		__F|=__D;
		Cycles=2; 
	}

	//************************************************************
	//                                                       
	//   Name    : PrcocessSKB                               
	//   Function: Skip a byte!                              
	//	 Opcodes : 80, 82, C2, E2, 04, 14, 34, 44, 54, 64, 74, D4, F4
	//                                                       
	//************************************************************
	final void ProcessSKB( )
	{
		__PC=(__PC+1)&0xffff;
		Cycles=3;
	}

	//************************************************************
	//                                                           *
	//   Name    : PrcocessSKW                                   *
	//   Function: Skip a word!                                  *
	//	 Opcodes : 0C, 1C, 3C, 5C, 7C, DC, FC					 *
	//                                                           *
	//************************************************************
	final void ProcessSKW( )
	{
		__PC=(__PC+2)&0xffff; 
		Cycles=4;
	}


	//****************************************************************************
	// Process the "STA" instruction 
	//****************************************************************************
	void ProcessSTA( ){

		int		address;
		int		zpage;

		switch (opcode) {
		   //case 0x85	:	zpage=GETABYTE(); Cycles=3; Poke(zpage,__A);        break;
		   //case 0x95	:	zpage=GETABYTE(); Cycles=4;Poke((zpage+__X)&255,__A);break;
		   //case 0x8D	:	address=GETAWORD(); Cycles=4; Poke(address,__A);    break;
		   //case 0x9D	:	address=GETAWORD(); Cycles=5; Poke(address+__X,__A); break;
		   //case 0x99	:	address=GETAWORD(); Cycles=5; Poke(address+__Y,__A); break;
		   case 0x81	:	zpage=GETABYTE(); Cycles=6; address=Deek((zpage+__X)&255); Poke(address,__A);    break;
		   case 0x91	:	zpage=GETABYTE(); Cycles=6; address=Deek(zpage); Poke(address+__Y,__A); break;
		}
	}

	//****************************************************************************
	// Process the "STX" instruction 
	//****************************************************************************
	void ProcessSTX( )
	{
		int		address;
		int		zpage;


		switch (opcode) {
		   case 0x86	:	zpage=GETABYTE(); 	Cycles=3; Poke(zpage,__X);      break;
		   case 0x96	:	zpage=GETABYTE(); 	Cycles=4; Poke((zpage+__Y)&0xff,__X);   break;
		   case 0x8E	:	address=GETAWORD(); 	Cycles=4; Poke(address,__X);    break;
		}
	}

	//****************************************************************************
	// Process the "STY" instruction 
	//****************************************************************************
	void ProcessSTY( )
	{
		int		address;
		int		zpage;

		switch (opcode) {
		   case 0x84	:	zpage=GETABYTE();   Cycles=3;  Poke(zpage,__Y);        	break;
		   case 0x94	:	zpage=GETABYTE();	  Cycles=4;  Poke((zpage+__X)&255,__Y); break;
		   case 0x8C	:	address=GETAWORD(); Cycles=4;  Poke(address,__Y);    		break;
		}
	}


	//****************************************************************************
	// Process the "TAX" instruction 
	//****************************************************************************
	final void ProcessTAX( )
	{
		__X=(__A&0xff);
		COMP(__X);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "TAY" instruction 
	//****************************************************************************
	final void ProcessTAY( )
	{
		__Y=(__A&0xff);
		COMP(__Y);
		Cycles=2; 
	}

	//****************************************************************************
	//
	// Process the "TSX" instruction 
	//
	//****************************************************************************
	final void ProcessTSX( )
	{
		__X=__SP&0xff;
		COMP (__X);
		Cycles=2; 
	}


	//****************************************************************************
	// Process the "TXA" instruction 
	//****************************************************************************
	final void ProcessTXA( )
	{
		__A=(__X&0xff);
		COMP (__A);
		Cycles=2; 
	}

	//****************************************************************************
	// Process the "TXS" instruction 
	//****************************************************************************
	final void ProcessTXS( )
	{
		__SP=(__X&0xff)|0x100;
		Cycles=2; 
	}

//****************************************************************************
// Process the "TYA" instruction 
//****************************************************************************
static void ProcessTYA( )
{
	__A=(__Y&0xff);
	COMP (__A);
	Cycles=2; 
}

//****************************************************************************
//
// Process the "AND" instruction 
//
//   XAA    ***
//   XAA transfers the contents of the X register to the A register and then 
//   ANDs the A register with an immediate value.
//
//   One supported mode:
//
//   XAA #ab         ;8B ab       ;No. Cycles= 2
//
//   Example:
//
//   XAA #$44        ;8B 44
//
//   Equivalent instructions:
//
//   TXA
//   AND #$44
//
//****************************************************************************
void ProcessXAA( )
{
	__A=__X&GETABYTE(); 
	Cycles=2;
	COMP (__A);
}


//****************************************************************************
// Process a single 6502 instruction
//****************************************************************************
void Process6502( void )
{
 	int	address=0;

/*	if(__PC == 0x100 )
	{
		___Asm {
			int 3
		};
	}
*/
	opcode=GETABYTE();

	int count = GetSizeOfInstruction(opcode);
	int PC = __PC;
	while(count!=0){
		ExecuteProfile[PC]++;
		if( MaxExecute<ExecuteProfile[PC]) MaxExecute = ExecuteProfile[PC];
		PC++;
		count--;
	}

	switch	(opcode){
		case	0x69:										
		case	0x65:										
		case	0x75:										
		case	0x6d:										
		case	0x7d:										
		case	0x79:										
		case	0x61:										
		case	0x71:	ProcessADC(); break;
		case	0x29:	
		case	0x25:	
		case	0x35:	
		case	0x2d:	
		case	0x3d:	
		case	0x39:	
		case	0x21:	
		case	0x31:	ProcessAND(); break;
		case	0x0a:
		case	0x06:
		case	0x16:
		case	0x0e:
		case	0x1e:	ProcessASL(); break;
		case	0x90:	ProcessBCC(); break;
		case	0xb0:	ProcessBCS(); break;
		case	0xf0:	ProcessBEQ(); break;
		case	0x24:
		case	0x2c:	ProcessBIT(); break;
		case	0x30:	ProcessBMI(); break;
		case	0xd0:	ProcessBNE(); break;
		case	0x10:	ProcessBPL(); break;
		case	0x00:	ProcessBRK(); break;
		case	0x50:	ProcessBVC(); break;
		case	0x70:	ProcessBVS(); break;
		case	0x18:	ProcessCLC(); break;
		case	0xd8:	ProcessCLD(); break;
		case	0x58:	ProcessCLI(); break;
		case	0xb8:	ProcessCLV(); break;
		case	0xc9:
		case	0xc5:
		case	0xd5:
		case	0xcd:
		case	0xdd:
		case	0xd9:
		case	0xc1:
		case	0xd1:	ProcessCMP(); break;
		case	0xe0:
		case	0xe4:
		case	0xec:	ProcessCPX(); break;
		case	0xc0:
		case	0xc4:
		case	0xcc:	ProcessCPY(); break;
		case	0xc6:
		case	0xd6:
		case	0xce:
		case	0xde:	ProcessDEC(); break;
		case	0xca:	__X=(__X-1)&0xff;	COMP(__X); Cycles=2; break;					// dex
		case	0x88:	__Y=(__Y-1)&0xff;	COMP(__Y); Cycles=2; break;					// dey
		case	0x49:
		case	0x45:
		case	0x55:
		case	0x4d:
		case	0x5d:
		case	0x59:
		case	0x41:
		case	0x51:	ProcessEOR(); break;
		case	0xe6:
		case	0xf6:
		case	0xee:
		case	0xfe:	ProcessINC(); break;
		case	0xe8:	__X=((__X+1)&0xff); COMP(__X); Cycles=2; break;					// inx
		case	0xc8:	__Y=((__Y+1)&0xff); COMP(__Y); Cycles=2; break;					// iny
		case	0x4c:
		case	0x6c:	ProcessJMP(); break;
		case	0x20:	ProcessJSR(); break;

		case	0xa9:	__A=GETABYTE(); Cycles=2; COMP(__A); break;						//lda #$00
		case	0xa5:	__A=Peek(GETABYTE()); Cycles=3; COMP(__A); break;					//lda $00
		case	0xad:	__A=Peek(GETAWORD()); Cycles=4; COMP(__A);  break;				//lda $0000
		case	0xb5:	__A=Peek( (GETABYTE()+__X) &0xff ); Cycles=4; COMP(__A); break;	//lda $00,x
		case	0xbd:
		case	0xb9:
		case	0xa1:
		case	0xb1:	ProcessLDA(); break;
		case	0xa2:	__X=GETABYTE(); Cycles=2; COMP(__X); break;						//ldx #$00
		case	0xa6:	__X=Peek(GETABYTE()); Cycles=3; COMP(__X); break;					//ldx $00
		case	0xae:	__X=Peek(GETAWORD()); Cycles=4; COMP(__X); break;					//ldx $0000
		case	0xb6:	__X=Peek((GETABYTE()+__Y)&0xff); Cycles=4; COMP(__X); break;		//ldx $00,y
		case	0xbe:	ProcessLDX(); break;
		case	0xa0:	__Y=GETABYTE(); Cycles=2; COMP(__Y); break;
		case	0xa4:	__Y=Peek(GETABYTE()); Cycles=3; COMP(__Y); break;
		case	0xac:	__Y=Peek(GETAWORD()); Cycles=4; COMP(__Y); break;
		case	0xb4:	__Y=Peek((GETABYTE()+__X)&255); Cycles=4; COMP(__Y); break;
		case	0xbc:	ProcessLDY(); break;
		case	0x4a:
		case	0x46:
		case	0x56:
		case	0x4e:
		case	0x5e:	ProcessLSR(); break;
		case	0xea:	Cycles=2; break;
		case	0x09:	__A|=GETABYTE(); Cycles=2; COMP(__A); break;					// ora #$00
		case	0x05:	__A|=Peek(GETABYTE()); Cycles=3; COMP(__A); break;			// ora $00
		case	0x15:
		case	0x0d:
		case	0x1d:
		case	0x19:
		case	0x01:
		case	0x11:	ProcessORA(); break;
		case	0x48:	PushReg (__A); Cycles=3; break;							// PHA
		case	0x08:	PushReg (__F); Cycles=3; break;							// PHP
		case	0x68:	__A=Pull();	COMP(__A);Cycles=4; break;					// PLA
		case	0x28:	__F=Pull();	Cycles=4; break;
		case	0x2a:
		case	0x26:
		case	0x36:
		case	0x2e:
		case	0x3e:	ProcessROL(); break;
		case	0x6a:
		case	0x66:
		case	0x76:
		case	0x6e:
		case	0x7e:	ProcessROR(); break;
		case	0x40:	ProcessRTI(); break;
		case	0x60:	ProcessRTS(); break;
		case	0xeb:
		case	0xe9:
		case	0xe5:
		case	0xf5:
		case	0xed:
		case	0xfd:
		case	0xf9:
		case	0xe1:
		case	0xf1:	ProcessSBC(); break;
		case	0x38:	__F|=__C;	Cycles=2;	break;
		case	0xf8:	__F|=__D;	Cycles=2;	break;
		case	0x78:	__F|=__I;	Cycles=2;	break;

		case	0x85:	Poke(GETABYTE(),__A); Cycles=3; break;				// sta
		case	0x95:	Poke((GETABYTE()+__X)&255,__A); Cycles=4; break;		// sta
		case	0x8D:	Poke(GETAWORD(),__A); Cycles=4; break;				// sta
		case	0x9d:	Poke(GETAWORD()+__X,__A); Cycles=5; break;
		case	0x99:	Poke(GETAWORD()+__Y,__A); Cycles=5; break;
		case	0x81:	Poke(Deek((GETABYTE()+__X)&255),__A);	Cycles=6; break;
		case	0x91:	Poke(Deek(GETABYTE())+__Y,__A);		Cycles=6; break;

		case	0x86:
		case	0x96:
		case	0x8e:	ProcessSTX(); break;
		case	0x84:
		case	0x94:
		case	0x8c:	ProcessSTY(); break;
		case	0xaa:	__X=(__A&0xff); COMP(__X); Cycles=2; break;
		case	0xa8:	__Y=(__A&0xff); COMP(__Y); Cycles=2; break;
		case	0xba:	__X=(__SP&0xff);COMP (__X); Cycles=2; break;
		case	0x8a:	__A=(__X&0xff); COMP (__A); Cycles=2; break;
		case	0x9a:	__SP=(__X&0xff)|0x100; Cycles=2; break;
		case	0x98:	__A=(__Y&0xff); COMP(__A); Cycles=2; break;

		//
		// Extended opcodes - these are hardly EVER used... so leave them unoptimised.
		//
		
		case	0x1a:	// NOP
		case	0x3a:
		case	0x5a:
		case	0x7a:
		case	0xda:
		case	0xfa:	Cycles=2; break;
		case	0x4b:	ProcessALR(); break;
		case	0x0f:
		case	0x1f:
		case	0x1b:
		case	0x07:
		case	0x17:
		case	0x03:
		case	0x13:	ProcessASO(); break;
		case	0x87:
		case	0x97:
		case	0x8f:
		case	0x83:	ProcessAXS(); break;
		case	0xc7:
		case	0xd7:
		case	0xcf:
		case	0xdf:
		case	0xdb:
		case	0xc3:
		case	0xd3:	ProcessDCM(); break;
		case	0xe7:
		case	0xf7:
		case	0xef:
		case	0xff:
		case	0xfb:
		case	0xe3:
		case	0xf3:	ProcessINS(); break;

		case	0xa7:	
		case	0xb7:
		case	0xaf:
		case	0xbf:
		case	0xa3:
		case	0xb3:	ProcessLAX(); break;

		case	0x47:
		case	0x57:
		case	0x4f:
		case	0x5f:
		case	0x5b:
		case	0x43:
		case	0x53:	ProcessLSE(); break;		
		case	0xab:	ProcessOAL(); break;
		case	0x27:
		case	0x37:
		case	0x2f:
		case	0x3f:
		case	0x3b:
		case	0x23:
		case	0x33:	ProcessRLA(); break;
		case	0x67:
		case	0x77:
		case	0x6f:
		case	0x7f:
		case	0x7b:
		case	0x63:
		case	0x73:	ProcessRRA(); break;
		case	0xcb:	ProcessSAX(); break;
		case	0x9c:	ProcessSAY(); break;
		case	0x80:
		case	0x82:
		case	0xc2:
		case	0xe2:
		case	0x04:
		case	0x14:
		case	0x34:
		case	0x44:
		case	0x54:
		case	0x64:
		case	0x74:
		case	0xd4:
		case	0xf4:	ProcessSKB(); break;
		case	0x89:	__PC=(__PC+1)&0xffff; Cycles=2; break;		// Another SKIP instrucion
		case	0x0c:
		case	0x1c:
		case	0x3c:
		case	0x5c:
		case	0x7c:
		case	0xdc:
		case	0xfc:	ProcessSKW(); break;
		case	0x8b:	ProcessXAA(); break;
		case	0x9f:
		case	0x93:	ProcessAXA(); break;
		case	0x0b:
		case	0x2b:	ProcessANC(); break;

		default:		ProcessHLT(); break;
	}
}

//****************************************************************************
// Reset CPU (softreset)
//****************************************************************************
void PatternFillMemory( void )
{
	int i;
	int count=0;
	for(i=0;i<16384;i++){
		pMemory[count]=0xff;
		pMemory[count+1]=0xff;
		pMemory[count+2]=0x00;
		pMemory[count+3]=0x00;
		count+=4;
	}
}

//****************************************************************************
// Reset CPU (softreset)
//****************************************************************************
void ResetPlus4( void )
{
   __PC=Deek(0xfffc)&0xffff;    // 0xf2a4;        // reset computer
   __SP=0x1ff;
   __A=0;
   __Y=0;
   __X=0;
   __F=0;
}
