#ifndef	__VIDEO_H__
#define	__VIDEO_H__

#include	<all.h>
#include	<memory.h>
#include	<ddraw.h>
#include	<d3d.h>
#include	<windows.h>

#define		__NT_DDRAW__							// Use DirectDraw Version 3 instead of DX6. (NT)


#define	MD_ERROR	1
#define	MD_OKAY		0


#define	VIDEO_WINDOWED			0xfffffffe			// open in a WINDOW
#define	VIDEO_PRIMARYDEVICE		1					// primary device number


// video mode options
#define	VIDEO_3D				0x00000001			// if not able to set, mode WILL fail
#define	VIDEO_VIDEORAM			0x00000002
#define	VIDEO_DOUBLEBUFFER		0x00000004			// Open screen in Double buffered mode (if fails fall back to single buffered mode)
#define	VIDEO_TRIPLEBUFFER		0x00000008			// Open screen in Triple buffered mode (if fails fall back to VIDEO_DOUBLEBUFFER)

// results
#define	VIDEO_SYS_VIDEORAM		0x80000000			// surface is in VIDEO ram
#define	VIDEO_SYS_PRIMARYLOST	0x40000000			// lost primary surface last call.
#define	VIDEO_SYS_BACKLOST		0x20000000			// lost back buffer last call.
#define	VIDEO_SYS_TRIPLEBUFFER	0x10000000			// system managed to create a triple buffer
#define	VIDEO_SYS_DOUBLEBUFFER	0x08000000			// system managed to create a triple buffer
#define	VIDEO_SYS_LOCKED		0x04000000			// Buffer is LOCKED




#ifndef	__NT_DDRAW__
//*****************************************************************************************************************
//*****************************************************************************************************************
//*****************************************************************************************************************
//*****************************************************************************************************************



// ******************************************************************
//
//	CScreenMode class
//
// ******************************************************************
class	CScreenMode
{
public:
		int32		m_ModeNumber;	// Made up screen mode number
		int32		m_Device;		// device number this is associated with.

		uint32		m_xsize;		// basic screen details
		uint32		m_ysize;
		uint32		m_bpp;	

		uint32		m_RedAND;		// Value to AND ARGB data with to maintain valid RED data
		uint32		m_RedShift;		// ROR value to move to correct position.
		uint32		m_GreenAND;		// Value to AND ARGB data with to maintain valid GREEN data
		uint32		m_GreenShift;	// ROR value to move to correct position.
		uint32		m_BlueAND;		// Value to AND ARGB data with to maintain valid BLUE data
		uint32		m_BlueShift;	// ROR value to move to correct position.
		
		// Hardly EVER used! usually "0"
		uint32		m_AlphaAND;		// Value to AND ARGB data with to maintain valid ALPHA data
		uint32		m_AlphaShift;	// ROR value to move to correct position.

class	CScreenMode	*m_pNext;		// next in list

		CScreenMode( LPDDSURFACEDESC2 lpDDSurfaceDesc );
};


// ******************************************************************
//
//	CDevice class
//
// ******************************************************************
class	CDevice			
{
public:
		int32			m_DeviceNumber;				// Device number 1=Primary Device, >=2 for all others
		char			*m_pDriverName;				// Name of driver (3DfxVoodoo, TNT)
		char			*m_pDriverDescription;		// Desc. "3Dfx Voodoo 2 dirver for Dx6" etc.
class	CDevice			*m_pNext;					// next in chain
class	CScreenMode		*m_pFirstMode;				// First Screen Mode on THIS device
class	CScreenMode		*m_pLastMode;				// Last mode in list

		// Under here, these may not exist 
		// on other systems.
		GUID		m_GUID;						// actual GUID
		GUID		*m_pGUID;					// pointer to GUID for THIS device
  
		CDevice( char *pName, char *pDesc );
		~CDevice( void );
};



// ******************************************************************
//
//	DirectDraw data
//
// ******************************************************************
typedef	struct	_tDDraw
{
	IDirectDraw				*pDD;				// pointer to a direct draw object
	IDirectDraw4			*pDD4;				// pointer to an upgraded direct draw 4 object
	DDSURFACEDESC2 			ddsd;				// Surface descriptor
	IDirectDrawSurface4		*pPrimary;			// primary surface pointer
	IDirectDrawSurface4		*pBack;				// Back buffer(s)
	IDirectDrawClipper		*pClipper;			// Clipper for VIDEO_WINDOWED mode

}tDDraw;

// ******************************************************************
//
//	CVideo class
//
// ******************************************************************
class	CVideo
{
public:
	uint32			m_flags;			// system flags
	int32			m_err;				// Last error message
	char*			m_pErr;				// IF an error, points to an error string

	HINSTANCE		m_hInstance;		// program instance
	CDevice			*m_pFirstDevice;	// First in the list of available devices
	CDevice			*m_pLastDevice;		// Last in the list of available devices
	
	int32			m_CurrentDevice;	// Current Device Number (ID)
	CDevice			*m_pCurrentDevice;	// pointer to current device class, or NULL
	int32			m_CurrentMode;		// Current screen mode number

	uint8			*m_pAddress;		// pointer to LOCKED screen buffer (or NULL)
	uint32			m_lPitch;				// line Pitch of screen (or 0)

	// Current mode details
	uint32			m_xsize;		// basic screen details
	uint32			m_ysize;
	uint32			m_bpp;			// Bit depth	
	uint32			m_RedAND;		// Value to AND ARGB data with to maintain valid RED data
	uint32			m_RedShift;		// ROR value to move to correct position.
	uint32			m_GreenAND;		// Value to AND ARGB data with to maintain valid GREEN data
	uint32			m_GreenShift;	// ROR value to move to correct position.
	uint32			m_BlueAND;		// Value to AND ARGB data with to maintain valid BLUE data
	uint32			m_BlueShift;	// ROR value to move to correct position.

	HWND			m_MainWindow;		// Window system is linked to.

	//
	// Data below here varies from system to system
	//
	tDDraw			m_DDraw;



	// Constructors and destructors.
	CVideo( HINSTANCE hInstance );
	~CVideo( void );
	
	// Main system calls
	char*			GetError( int hResult );
	int32			Init( uint32 flags = 0 );
	int32			CheckMode( uint32 width, uint32 height, uint32 bpp = 16 );
	CScreenMode*	FindMode( int32 ModeNumber );
	CDevice*		FindDevice( int32 deviceid );							//
	int				SetDevice( int32 device );								// set current device, also close any open devices and screen modes
	int				CloseScreen( void );									// close current mode (windowed or fullscreen)
	int				SetMode( HWND Window, int32 Mode = VIDEO_WINDOWED, int32 flags = 0 );
	int				Flip( void );
	int				ClearScreen( uint32 ARGB, int x1=1, int y1=0, int x2=0, int y2=0);
	void			ShutDown( void );
	int32			Lock( void );
	int32			Unlock( void );
	int32			SetGamma( float RedGamma, float GreenGamma, float BlueGamma );

	friend	BOOL FAR PASCAL EnumDrivers_callback(GUID FAR* lpGUID, LPSTR lpDriverDesc, LPSTR lpDriverName, LPVOID lpContext );


private:
	void	ReleaseDDrawHandle( void );
	int		GetDDrawHandle( CDevice *pDevice );
	int		OpenWindowedMode( HWND Window, int32 Mode, int32 flags );
	int		OpenFullScreenMode( HWND Window, int32 Mode, int32 flags );
	void	ComputeGammaCorrectedLUTentries(uint16 *pLUT, float Gamma);

};
#else
//*****************************************************************************************************************
//*****************************************************************************************************************
//
//  NT version (DirectDraw 3)
//
//*****************************************************************************************************************
//*****************************************************************************************************************




// ******************************************************************
//
//	CScreenMode class
//
// ******************************************************************
class	CScreenMode
{
public:
		int32		m_ModeNumber;	// Made up screen mode number
		int32		m_Device;		// device number this is associated with.

		uint32		m_xsize;		// basic screen details
		uint32		m_ysize;
		uint32		m_bpp;	

		uint32		m_RedAND;		// Value to AND ARGB data with to maintain valid RED data
		uint32		m_RedShift;		// ROR value to move to correct position.
		uint32		m_GreenAND;		// Value to AND ARGB data with to maintain valid GREEN data
		uint32		m_GreenShift;	// ROR value to move to correct position.
		uint32		m_BlueAND;		// Value to AND ARGB data with to maintain valid BLUE data
		uint32		m_BlueShift;	// ROR value to move to correct position.
		
		// Hardly EVER used! usually "0"
		uint32		m_AlphaAND;		// Value to AND ARGB data with to maintain valid ALPHA data
		uint32		m_AlphaShift;	// ROR value to move to correct position.

class	CScreenMode	*m_pNext;		// next in list

		CScreenMode( LPDDSURFACEDESC lpDDSurfaceDesc );
};


// ******************************************************************
//
//	CDevice class
//
// ******************************************************************
class	CDevice			
{
public:
		int32			m_DeviceNumber;				// Device number 1=Primary Device, >=2 for all others
		char			*m_pDriverName;				// Name of driver (3DfxVoodoo, TNT)
		char			*m_pDriverDescription;		// Desc. "3Dfx Voodoo 2 dirver for Dx6" etc.
class	CDevice			*m_pNext;					// next in chain
class	CScreenMode		*m_pFirstMode;				// First Screen Mode on THIS device
class	CScreenMode		*m_pLastMode;				// Last mode in list

		// Under here, these may not exist 
		// on other systems.
		GUID		m_GUID;						// actual GUID
		GUID		*m_pGUID;					// pointer to GUID for THIS device
  
		CDevice( char *pName, char *pDesc );
		~CDevice( void );
};



// ******************************************************************
//
//	DirectDraw data
//
// ******************************************************************
typedef	struct	_tDDraw
{
	IDirectDraw				*pDD;				// pointer to a direct draw object
//	IDirectDraw4			*pDD4;				// pointer to an upgraded direct draw 4 object
	DDSURFACEDESC 			ddsd;				// Surface descriptor
	IDirectDrawSurface		*pPrimary;			// primary surface pointer
	IDirectDrawSurface		*pBack;				// Back buffer(s)
	IDirectDrawClipper		*pClipper;			// Clipper for VIDEO_WINDOWED mode

}tDDraw;

// ******************************************************************
//
//	CVideo class
//
// ******************************************************************
class	CVideo
{
public:
	uint32			m_flags;			// system flags
	int32			m_err;				// Last error message
	char*			m_pErr;				// IF an error, points to an error string

	HINSTANCE		m_hInstance;		// program instance
	CDevice			*m_pFirstDevice;	// First in the list of available devices
	CDevice			*m_pLastDevice;		// Last in the list of available devices
	
	int32			m_CurrentDevice;	// Current Device Number (ID)
	CDevice			*m_pCurrentDevice;	// pointer to current device class, or NULL
	int32			m_CurrentMode;		// Current screen mode number

	uint8			*m_pAddress;		// pointer to LOCKED screen buffer (or NULL)
	uint32			m_lPitch;				// line Pitch of screen (or 0)

	// Current mode details
	uint32			m_xsize;		// basic screen details
	uint32			m_ysize;
	uint32			m_bpp;			// Bit depth	
	uint32			m_RedAND;		// Value to AND ARGB data with to maintain valid RED data
	uint32			m_RedShift;		// ROR value to move to correct position.
	uint32			m_GreenAND;		// Value to AND ARGB data with to maintain valid GREEN data
	uint32			m_GreenShift;	// ROR value to move to correct position.
	uint32			m_BlueAND;		// Value to AND ARGB data with to maintain valid BLUE data
	uint32			m_BlueShift;	// ROR value to move to correct position.

	HWND			m_MainWindow;		// Window system is linked to.

	//
	// Data below here varies from system to system
	//
	tDDraw			m_DDraw;



	// Constructors and destructors.
	CVideo( HINSTANCE hInstance );
	~CVideo( void );
	
	// Main system calls
	char*			GetError( int hResult );
	int32			Init( uint32 flags = 0 );
	int32			CheckMode( uint32 width, uint32 height, uint32 bpp = 16 );
	CScreenMode*	FindMode( int32 ModeNumber );
	CDevice*		FindDevice( int32 deviceid );							//
	int				SetDevice( int32 device );								// set current device, also close any open devices and screen modes
	int				CloseScreen( void );									// close current mode (windowed or fullscreen)
	int				SetMode( HWND Window, int32 Mode = VIDEO_WINDOWED, int32 flags = 0 );
	int				Flip( void );
	int				ClearScreen( uint32 ARGB, int x1=1, int y1=0, int x2=0, int y2=0);
	void			ShutDown( void );
	int32			Lock( void );
	int32			Unlock( void );
	//int32			SetGamma( float RedGamma, float GreenGamma, float BlueGamma );

	friend	BOOL FAR PASCAL EnumDrivers_callback(GUID FAR* lpGUID, LPSTR lpDriverDesc, LPSTR lpDriverName, LPVOID lpContext );


private:
	void	ReleaseDDrawHandle( void );
	int		GetDDrawHandle( CDevice *pDevice );
	int		OpenWindowedMode( HWND Window, int32 Mode, int32 flags );
	int		OpenFullScreenMode( HWND Window, int32 Mode, int32 flags );
};


//*****************************************************************************************************************
//*****************************************************************************************************************
//*****************************************************************************************************************
//*****************************************************************************************************************

#endif
extern	int	Len( char *pStr );		// Basic "C" function to return the length of a string



#endif //__VIDEO_H__