#ifndef __SYMBOL__H_
#define	__SYMBOL__H_

#define	SYMBOL_BLOCK_COUNT	(1024*512)	// allocate 512K at once for symbols and strings
#define	SYMBOL_LEN			128			// max string used in symbol

#define	DEFINED_YET		0x80000000
#define	TYPE_MASK		0xff00
#define	TYPE_LABEL		0x0100
#define	TYPE_SET		0x0200
#define	TYPE_RS			0x0300
#define	TYPE_EQU		0x0400
#define	TYPE_MACRO		0x0800
#define	TYPE_MACROPARAM	0x1000			// macro has paramaters

//
// Our basic symbol structure
//
typedef	struct	tSym{
		uint32	dwFlags;				// %ffffffffffffffffffffffffSSSSSSSS	( f=flag, s=tring size (length) )
		int32	Value;					// value
		int32	Value2;					// extra user value
		
		uint32	length;					// string length
		char	*pString;				// string assosiated with symbol
		char	*pFileName;// Name of file this symbol was defined in.
		uint32	LineNumber;				// Line number symbol was defined at.
		char	Name[SYMBOL_LEN+1];		// up to SYMBOL_LEN (63 last check) characters a symbol
struct	tSym	*pLeft;					// left branch of binary tree.
struct	tSym	*pRight;
}tSym, *PSym;
//
// A linked list of symbol space
//
typedef	struct	tSymSpace{
		uint8		*pMemory;		// pointer to Current block
		uint32		dwSize;			// size of block
struct	tSymSpace	*pNext;			// pointer to next "block"
}tSymSpace, *PSymSpace;


extern	PSym		pLastSymbol;			// Last add symbol location (even if failed due to redefinition)

extern	uint32		dwSymbolCount;
extern	uint32		dwSymbolMemoryUsed;
extern	int32		g_MacroCounter;				// auto inc counter to allow local labels inside macros
extern	char		pszLastLabel[4096];			// pointer to LAST normal label (used to append locals)
extern	char		g_LastLastSymbol[4096];		// previous symbol before this one....


void	InitSymbols	( void );
void*	AllocSpace	( unsigned int size=0 );
PSym	FindSymbol	( char *pszName );
PSym	AddSymbol	( char *pszName, int32 Value, int32 Value2, uint32 flags );
PSym	AddNode		( PSym pSym);
void	AddRecursively( PSym pSym );		// used in the delete
void	DeleteSymbol( char *pszName );
void DumpSymbolTable( char* pOutputName );
void FreeSymbols( void );
char*	FindValue( int value );



#endif	//__SYMBOL__H_