//******************************************************************************
//
//								  Windows Interface
//
//							 Copyright Mike Dailly 1996
//
//
//******************************************************************************


#ifndef __WindowInterface__
#define __WindowInterface__


//****************************************************************************
//
//	Systems Mouse structure
//
//****************************************************************************

typedef	struct	tMouse {
			uint8		*pHand;					// standard mouse pointer (non-clicked)
			int16		clickx,clicky;			// click pixel offsets
			int16		xsize,ysize;			// x and y size of mouse
			int16		xoffset,yoffset;		// drawing offset (for centering)
}tMouse;


//****************************************************************************
//
//	Windows FONT structure.
//
//****************************************************************************

typedef	struct	tFont {
			uint8		*pBitMap;				// actual font data
			int16		*pTable;					// proportional indexs.
			int16		xsize,ysize;			// font x and y size
}tFont;


//****************************************************************************
//
// Gadget structures and defines
//
//****************************************************************************

#define	BOOLEAN			1					// boolean type on/off (up/down box)
#define	TICK				2					// boolean type on/off (tick mark)
#define	HPROP				3					// horizontal Slider type
#define	VPROP				4					// vertical Slider type
#define	HVPROP			5					// 2d slider
#define	VHPROP			5					// 2d slider
#define	STRING			6					// Text type (any char)
#define	INTEGER			7					// interger type (numbers only)
#define	TEXT				8					// text string. ie print ??? at x,y
#define	TEXTBUTTON		9					// Text Button. on/off state

#define	gACTIVE			1					// gadget active?
#define	gDRAWN			2					// gadget to be drawn? (invisable)
#define	gCOLOUR			4					// gadget has its own colour scheme
#define	gREFRESH			0x80000000		// gadget needs refreshing

typedef	struct	tGadget{
			int16		x,y;
			int16		xsize,ysize;			// used for prop gadgets
			int16		maxxsize,maxysize;	// also used for string lengths	
			int16		minxsize,minysize;
			int32		type;						// gadget type
			int32		flags;					// gadget flags
			int32		value;					// integer value. 
													// Also true/false value. 
													// And used in Cycle gadgets!
		 	uint8		PenColour;				// pen colours
		 	uint8		DrawCols[5];			// pen colours for gadget (for override)
			char		*pMessage;				// message for char type
			int32		misc[20];				// misc storage
			

//			// how to DRAW the gadget - to be added
//			int16		nxsize,nysize;			// graphic x and y size (non-selected)
//			int16		sxsize,sysize;			// graphic x and y size (selected)
//			int32		nrealwidth;				// actual width of data
//			int32		srealwidth;				// actual width of data
//			uint8		*pNBitMap;				// bitmap address (non-selected)
//			uint8		*pSBitMap;				// bitmap address (non-selected)

struct	tGadget	*pLastGadget;
struct	tGadget	*pNextGadget;
}tGadget;


//****************************************************************************
//
// Menu structures
//
//****************************************************************************

#define	mHIDE					1				// skip item
#define	mDISABLE				2				// ghost item.
#define	mTICK   				4				// tick menu item
#define	mRIGHTALIGN			8				// ie Help option
#define	mSELECTED 			0x10000000	// show child level
#define	mCHOICE   			0x20000000	// This is the current choice!

typedef	struct	tMenu {
		 int32		Options;					// system status flags
		 int32		flags;					// system status flags
		 char			*pMessage;				// menu text
		 int32		ID;						// menu ID. this is what returns when item is selected


struct tMenu		*pChildMenu;			// pointer to a child menu
struct tMenu		*pNextItem;				// next menu item (lists down)
}tMenu;


//****************************************************************************
//
// Window structures and defines
//
//****************************************************************************

//
// window options
//
#define	NONE						0					// window options
#define	DRAGDROP					0x00000001
#define	HDRAGDROP				0x00000002		// allow Horizontal drag drop
#define	VDRAGDROP				0x00000004		// allow Vorizontal drag drop
#define	RESIZE					0x00000008		// resize gadget
#define	CLOSE						0x00000010		// close gadget
#define	DRAGBAR					0x00000020		// dragbar 
#define	BORDERS					0x00000040		// borders (and gadgets)
#define	TRANSPARENT				0x20000000		// see through
#define	HIDE						0x40000000		// hide window (dont process)


//
// system flags
//
#define	wSELECTED		1					// window is currently active
#define	wMOVING			2					// window is being moved
#define	wRESIZE			4					// window is being resized
#define	wACTIVE   		0x02000000		// Mouse is over THIS window
#define	wPRINTMENU		0x04000000		// menu is being printed (no selection yet)
#define	wMENU   			0x08000000		// menu item has been selected
#define	wACTION 			0x10000000		// window needs processing (something happened)
#define	wTRANSPARENT	0x20000000		// window "flickers" so you can see through it
#define	wCLOSE			0x40000000		// window close request recieved
#define	wREFRESH			0x80000000		// window needs refreshing



typedef	struct tWindow{
		 int16		x,y;											// window X and Y cords. 
		 int16		xsize,ysize;								// current X and Y size
		 int16		maxxsize,maxysize;						// MAX x and y sizes
		 int16		clipx1,clipy1,clipx2,clipy2;			// window clip regions
		 int16		minx,miny,maxx,maxy;						// Min+Max (x,y) cords
		 int32		offsetx,offsety;							// offset from topleft
		 int32		vxsize,vysize;								// virtual x and y sizes
		 int32		vscrollx,vscrolly;						// virtual amounts for x and y
		 uint32		flags;										// misc system flags
		 uint32		Options;										// Window options
		 uint32		MenuID;										// MenuID (when selected)
		 uint8		WindowColour;								// window fill colour
		 uint8		PenColour;								   // Colour for TEXT etc..
		 uint8		DrawCols[5];								// pen colours for window
		 uint8		*BitMap;
		 uint8		*pBackDrop;									// backdrop (rather than colour)
		 tFont		*pFont;										// pointer to fonts
struct tMenu		*pFirstMenu;								// point to the first menu item
struct tGadget		*pFirstGadget;								// pointer to FIRST gadget
struct tGadget		*pLastGadget;								// pointer to LAST  gadget
struct tWindow		*pParentWindow;						
struct tWindow		*pFirstChildWindow;						// child windows
struct tWindow		*pLastChildWindow;						
struct tWindow		*pNextWindow;
struct tWindow		*pLastWindow;
}tWindow;

//******************************************************************************
//
//	Actual system calls and global variables
//
//******************************************************************************

#ifndef __WindowInterCalls__
#define __WindowInterCalls__


			extern	tWindow	*pScreen;
			extern	tFont		*pFont;
			extern	tFont		*pFont2;
			extern	uint8		EnableMouse;
			extern	tMouse	*pMouse;
			extern 	uint32	FileSize;
			extern	uint8		DrawCols[5];		// current DRAWING colours


//
// Main windows calls
//
extern	void		InitSystem ( uint8 *Scr, int32 xsize, int32 ysize );
extern	void 		ShutDownWindows( void );
extern	void 		ProcessInterface ( char currentkey, int16 mx, int16 my, int16 mbut );
extern	void		RenderInterface ( void );
extern	uint32	GetFileSize (char *name);
extern	uint8 	*WindLoad(char *name);
extern	void		Save(char *name, uint8 *Data, uint32 Size);
extern	void		SetColour (int ColR, int R, int G, int B);



// general window function calls	 
extern	void 		SetDefaultPen 	   (uint8 wh, uint8 lg, uint8 mg, uint8 dg, uint8 bl);
extern	void 		SetWindowPen 		(tWindow *pWindow, uint8 wh, uint8 lg, uint8 mg, uint8 dg, uint8 bl);
extern	void 		SetGadgetPen 		(tGadget *pgadget, uint8 wh, uint8 lg, uint8 mg, uint8 dg, uint8 bl);
extern	void 		Print 		 		(tWindow *pWindow, int16 x1,int16 y1, char *pMessage);
extern	tWindow *MessageBox 			(tWindow *pWindow, int16 x,int16 y,char *pMessage);
extern	tWindow *OpenWindow  		(tWindow *pParent, int16 x,int16 y,int16 xsize,int16 ysize);
extern	void 	   CloseWindow 		(tWindow *pWindow);
extern	tGadget *AddBooleanGadget 	(tWindow *pWindow, int32 type, int16 x, int16 y);
extern	tGadget *AddPropGadget		(tWindow *pWindow, int16 x, int16 y,
												 int32 dispxsize, int32 minx, int32 maxx,
												 int32 dispysize, int32 miny, int32 maxy);
extern	tGadget *AddStringGadget 	(tWindow *pWindow, int32 type, int16 x, int16 y, 
												 int32	gxsize,int32 maxsize);
extern	tGadget *AddText 			 	(tWindow *pWindow, int16 x, int16 y,char *pMessage);
extern	tGadget *AddTextButton 	 	(tWindow *pWindow, int16 x, int16 y,
											  	 int16 xs,int16 ys,  char *pMessage);
extern	tMenu	  *AddMenuItem 		(tWindow *pWindow ,tMenu *pMasterMenu, char *pMessage, uint32 ID);
extern	void		ResetMouse ( void );
extern	void		SetMouse ( tMouse *MouseData );


// lower level system calls
extern	void	PrintWindowFrame	(tWindow *pWindow);
extern	void	RefreshWindow		(tWindow *pWindow);
extern	void	LimitWindow			(tWindow *pWB, tWindow *pWindow);
extern	void	SetWindowLimits	(tWindow *pWindow,int16 x,int16 y,int16 x2,int16 y2);
extern	void	SetDefaultColours	( void );



#endif	// calls


#ifndef __DMAinterfaceASMCalls__
#define __DMAinterfaceASMCalls__

extern	uint8	*FontData;
extern	uint8	*GadData;

extern	void cdecl SysBlock     ( tWindow *pWindow, uint8 *data, int32 x1, int32 y1, int32 xs,int32 ys,int32 realxsize);
extern	void cdecl SysSprite    ( tWindow *pWindow, uint8 *data, int32 x1, int32 y1, int32 xs,int32 ys,int32 realxsize);
extern	void cdecl SysBlockMono ( tWindow *pWindow, uint8 *data, int32 x1, int32 y1, int32 xs,int32 ys,int32 realxsize, uint8 fcol, uint8 bcol);
extern	void cdecl SysSpriteMono( tWindow *pWindow, uint8 *data, int32 x1, int32 y1, int32 xs,int32 ys,int32 realxsize, uint8 col);


#endif	// calls


#endif	// #defines
