#include <cplus4.h>
#include <sound.h>
#include <myprocess.h>
#include <machine.h>


extern	int pROM[];
extern	int	ROM3PLUS1[];

// ******************************************************************
//
// Create plus/4 hardware
//
// ******************************************************************
CPlus4::CPlus4( void )
{
	Init();
}

// ******************************************************************
//
// Delete plus4 hardware
//
// ******************************************************************
CPlus4::~CPlus4( void )
{
}

// ******************************************************************
//
// Init machine
//
// ******************************************************************
int CPlus4::Init( void )
{
	int	i;
	for (i=0;i<65546;i++){
		m_pMemory[i]=0;
	}

	m_skipit=0;
	m_ROM=1;
	m_ROMACTIVE=1;
	m_Keys[0]=m_Keys[1]=m_Keys[2]=m_Keys[3]=m_Keys[4]=m_Keys[5]=m_Keys[6]=m_Keys[7]=m_Keys[8]=m_Keys[9]=m_Keys[10]=0xff;


	// init basic ROM setup
	m_pRom0 = &pROM[0];
	m_pRom1 = &ROM3PLUS1[0];
	m_pRom2 = &m_ROM3[0];
	m_pRom3 = &m_ROM4[0];
	m_pRom4 = &pROM[16384];
	m_pRom5 = &ROM3PLUS1[16384];
	m_pRom6 = &m_ROM7[0];
	m_pRom7 = &m_ROM8[0];

	return 0;
}

// ******************************************************************
//
// Get byte from memory
//
// ******************************************************************
int CPlus4::Peek(int address)
{
	if((m_pMemory[0xfd21]&0xc0) ==0 ){
		m_pMemory[0xfd21] ^= 0xc0;
	}
	address&=0xffff;
	// IS rom active? if not... read from RAM
	if ((m_ROMACTIVE==0) || (address<0x8000)) {
		return m_pMemory[address];
	}else{
		// ALL hardware and TED access via MEMORY
		if( ( address >=0xfd10 ) && (address < 0xff40 ) ){
			switch( address ){
				case	0xfef1: { 
									int	a;
									a = m_pMemory[ 0xfef1 ];
									m_pMemory[ 0xfef1 ] = 0;
									return a;
								}			
				case	0xff15: 
				case	0xff16: 
				case	0xff17: 
				case	0xff18: 
				case	0xff19: return m_pMemory[address] | 0x80; break;

				default: return m_pMemory[ address ];
			}
		}
		else if( ( address>=0xfc00 ) && ( address<=0xfcff ) ){		// this ALWAYS comes fom main rom.
			return m_pROM[address&0x7fff];
		}
		else if( address<0xc000 ){
			return m_pRomLo[address&0x3fff];
		}
		else{
			return m_pRomHi[address&0x3fff];
		}
	}
}

//****************************************************************************
// DPEEK memory
//****************************************************************************
int CPlus4::Deek(int address)
{
	int	a;
	a = Peek(address);
	a |= (Peek(address+1)<<8);
	return a;
}


//****************************************************************************
// Poke memory - stores into memory, but does NOT process hardware.
//****************************************************************************
void CPlus4::POKE (int address,int val)
{
	address&=0xffff;
	val&=0xff;

	m_pMemory[address]=val;
	if ((address>=0xff00) && (address<=0xff40)) {
		pROM[address&0x7fff]=val;
	}
}


void CPlus4::DOKE (int address,int val)
{
	POKE (address,val&0xff);
	POKE (address+1,(val>>8)&0xff);
}

//****************************************************************************
// Small Hardware functions
//****************************************************************************

// Read the keybaord
int CPlus4::ReadKeyboard( int a )
{
	int	b=0xff;

	if ((a&1)==0) b&=m_Keys[1];
	if ((a&2)==0) b&=m_Keys[2];
	if ((a&4)==0) b&=m_Keys[3];
	if ((a&8)==0) b&=m_Keys[4];
	if ((a&16)==0) b&=m_Keys[5];
	if ((a&32)==0) b&=m_Keys[6];
	if ((a&64)==0) b&=m_Keys[7];
	if ((a&128)==0) b&=m_Keys[8];

	return b;
}

// Timer 1
void CPlus4::SetTimer1LO(int val)
{
	pROM[0x7f00]=pROM[0x7f01]=m_pMemory[0xff00]=m_pMemory[0xff01]=0;		// clear timer for CPU reads
	m_TIMER1_ACTIVE=255;						// disable timer 1
	m_Timer1Value=val<<16;					// set LO of timer 1
}
void CPlus4::SetTimer1(int val)
{
	m_TIMER1_ACTIVE=0; 
	m_Timer1=m_Timer1Value=((m_Timer1Value&0xff0000)|val<<24); 
	pROM[0x7f00]=m_pMemory[0xff00]=((m_Timer1>>16)&0xff);    // write NEW timer LOW
	pROM[0x7f01]=m_pMemory[0xff01]=((m_Timer1>>24)&0xff);    // write NEW timer HIGH
}

// Timer 2
void CPlus4::SetTimer2LO(int val)
{
	pROM[0x7f03]=pROM[0x7f02]=m_pMemory[0xff02]=m_pMemory[0xff03]=0;		// clear timer for CPU reads
	m_TIMER2_ACTIVE=255;						// disable timer 2
	m_Timer2=val<<16;							// set LO of timer 2
}
void CPlus4::SetTimer2(int val)
{
	m_TIMER2_ACTIVE=0; 
	m_Timer2=m_Timer1Value=((m_Timer2&0xff0000)|val<<24); 
	pROM[0x7f02]=m_pMemory[0xff02]=((m_Timer2>>16)&0xff);    // write NEW timer LOW
	pROM[0x7f03]=m_pMemory[0xff03]=((m_Timer2>>24)&0xff);    // write NEW timer HIGH
}

// Timer 3
void CPlus4::SetTimer3LO(int val)
{
	pROM[0x7f05]=pROM[0x7f04]=m_pMemory[0xff04]=m_pMemory[0xff05]=0;		// clear timer for CPU reads
	m_TIMER3_ACTIVE=255;						// disable timer 3
	m_Timer3=val<<16;							// set LO of timer 3
}
void CPlus4::SetTimer3(int val)
{
	m_TIMER3_ACTIVE=0; 
	m_Timer3=m_Timer1Value=((m_Timer3&0xff0000)|val<<24); 
	pROM[0x7f04]=m_pMemory[0xff04]=((m_Timer3>>16)&0xff);    // write NEW timer LOW
	pROM[0x7f05]=m_pMemory[0xff05]=((m_Timer3>>24)&0xff);    // write NEW timer HIGH
}

//****************************************************************************
// Poke memory
//****************************************************************************
#define	SIDMIRROR		(0xfe80-0xfd40)

void CPlus4::Poke(int address,int val)
{
	address&=0xffff;
	val&=0xff;

//	if(address==0xfd00){
//		address&=0xffff;
//	}

	m_pMemory[address]=val;
	if((address>=0xfe80) && (address<0xfe99)){
		address -= (0xfe80-0xfd40);
	}
	if((address>=0xfdd0) && (address<=0xfddf)){
		switch(address){
			case	0xfdd0:	m_pRomLo = m_pRom0;	m_pRomHi = m_pRom4; return;
			case	0xfdd1:	m_pRomLo = m_pRom1;	m_pRomHi = m_pRom4; return;
			case	0xfdd2:	m_pRomLo = m_pRom2;	m_pRomHi = m_pRom4; return;
			case	0xfdd3:	m_pRomLo = m_pRom3;	m_pRomHi = m_pRom4; return;
			case	0xfdd4:	m_pRomLo = m_pRom0;	m_pRomHi = m_pRom5; return;
			case	0xfdd5:	m_pRomLo = m_pRom1;	m_pRomHi = m_pRom5; return;
			case	0xfdd6:	m_pRomLo = m_pRom2;	m_pRomHi = m_pRom5; return;
			case	0xfdd7:	m_pRomLo = m_pRom3;	m_pRomHi = m_pRom5; return;
			case	0xfdd8:	m_pRomLo = m_pRom0;	m_pRomHi = m_pRom6; return;
			case	0xfdd9:	m_pRomLo = m_pRom1;	m_pRomHi = m_pRom6; return;
			case	0xfdda:	m_pRomLo = m_pRom2;	m_pRomHi = m_pRom6; return;
			case	0xfddb:	m_pRomLo = m_pRom3;	m_pRomHi = m_pRom6; return;
			case	0xfddc:	m_pRomLo = m_pRom0;	m_pRomHi = m_pRom7; return;
			case	0xfddd:	m_pRomLo = m_pRom1;	m_pRomHi = m_pRom7; return;
			case	0xfdde:	m_pRomLo = m_pRom2;	m_pRomHi = m_pRom7; return;
			case	0xfddf:	m_pRomLo = m_pRom3;	m_pRomHi = m_pRom7; return;
		}
		return;
	}
	else if((address>=0xfd20) && (address<0xfd40)) {
		switch( address ){
			case	0xfd20:	m_pMemory[0xfd20] = val; break;
			case	0xfd21:	m_pMemory[0xfd21] = 0xc0|val; break;
			case	0xfd22:	m_pMemory[0xfd22] = val; m_pMemory[0xfd21] = 0; break;
			case	0xfd30:
			case	0xfd31:
			case	0xfd32:
			case	0xfd33:
			case	0xfd34:
			case	0xfd35:
			case	0xfd36:
			case	0xfd37:
			case	0xfd38:
			case	0xfd39:
			case	0xfd3a:
			case	0xfd3b:
			case	0xfd3c:
			case	0xfd3d:
			case	0xfd3e:
			case	0xfd3f:	m_pMemory[0xfd30] = val; m_pMemory[address] = val;
			default:
				m_pMemory[address] = val;
		}
	}
	else if((address>=0xfd40) && (address<0xfd59)) {

		switch (address){
			// Channel 1
			case	0xfd40:	
			case	0xfd41:	
			case	0xfd42:	
			case	0xfd43:	
			case	0xfd44:	
			case	0xfd45:	
			case	0xfd46:	m_pRomHi[address&0x3fff] = (m_pMemory[address+SIDMIRROR] = (m_pMemory[address]=val)); 
							InitChannel1(address,val);
							break;

			// Channel 2
			case	0xfd47:	
			case	0xfd48:	
			case	0xfd49:	
			case	0xfd4a:	
			case	0xfd4b:	
			case	0xfd4c:	
			case	0xfd4d:	m_pRomHi[address&0x3fff] = (m_pMemory[address+SIDMIRROR] = (m_pMemory[address]=val)); 
							InitChannel2(address,val);
							break;				
			
			// Channel 3
			case	0xfd4e:	
			case	0xfd4f:	
			case	0xfd50:	
			case	0xfd51:	
			case	0xfd52:	
			case	0xfd53:	
			case	0xfd54:	m_pRomHi[address&0x3fff] = (m_pMemory[address+SIDMIRROR] = (m_pMemory[address]=val)); 
							InitChannel3(address,val);
							break;				


			// Just store case....Control registers, processed each "Click"
			case	0xfd55:	
			case	0xfd56:	
			case	0xfd57:	
			case	0xfd58:	m_pRomHi[address&0x3fff] = (m_pMemory[address+SIDMIRROR] = (m_pMemory[address]=val)); 
							WriteChannelData(address,val);			
							break;
		}
	}
	else if((address>=0xfef0) && (address<0xff40)) {

		//romaddress=address&0x7fff;
		//pROM[romaddress]=val;

		switch (address){
			// Timers
			case	0xfef0:	ProcessDiskDrive ( val ); break;
			case	0xff00:	m_pMemory[address]=val; SetTimer1LO(val); break;
			case	0xff01:	m_pMemory[address]=val; SetTimer1(val);	break;
			case	0xff02:	m_pMemory[address]=val; SetTimer2LO(val); break;
			case	0xff03:	m_pMemory[address]=val; SetTimer2(val);	break;
			case	0xff04:	m_pMemory[address]=val; SetTimer3LO(val); break;
			case	0xff05:	m_pMemory[address]=val; SetTimer3(val);	break;

			case	0xff08: pROM[0x7f08]=(m_pMemory[0xff08]=ReadKeyboard(val)); break; 
			case	0xff09: pROM[0x7f09]=(m_pMemory[0xff09]&=~val); break; 
			case	0xff1d: m_pMemory[0xff1d] = val; ResetScreen(); break;
			case	0xff1f:	pROM[0x7f1f]=val; m_pMemory[address]=val; break; 
			case	0xff3e:	m_ROMACTIVE=1; pROM[0x7f3e] = m_pMemory[address]=val; break; 
			case	0xff3f:	m_ROMACTIVE=0; pROM[0x7f3f] = m_pMemory[address]=val; break; 
			default:{
					m_pMemory[address]=val;
					pROM[address&0x7fff]=val;		// never 
					break;
			}
		}
	}
	else if( address == 0x1 ){
		m_pMemory[ 0x0001 ] = val | ((val&0x4)<<5);
	}
}
