//******************************************************************************
//
//								Profiler.C
//
//									By
//
//						        Mike Dailly
//
//			   Copyright (c). 1999-2007 All Rights reserved
//
//
//******************************************************************************
#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#define	_WIN32_WINNT	0x0501

#include	<conio.h>
#include	<windows.h>
#include	<Wingdi.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>

#define __WINDOWS_HEADERS__
#include "headers\wintest.h"			// define #defines etc..
#include "resource.h"
#include "cvideo.h"
#include "csound.h"
#include "main.h"
#include "machine.h"
#include "6502.h"
#include "myprocess.h"
#include "sound.h"
#include <mmsystem.h>
#include "cdirectinput.h"
#include "debug.h"
#include "symbol.h"
#include <sys/stat.h>


extern	void Hex (int number,int count);
extern	HWND	g_ProfileWindow;
extern	char	FileName[512];

#define	WINDOW_WIDTH	800
#define	WINDOW_HEIGHT	320

	enum	eProfileMode
	{
		ReadOnly=0,
		WriteOnly,
		ExecuteOnly,
		ReadWrite,
		ReadWriteExecute
	};
	enum	eBarMode
	{
		Accumulate=0,
		Average,
		Maximum,
		Minimum
	};

	WNDPROC		WNProc;
	HWND		g_hDlg;
	HDC			g_hDC;
	float		ProfileStart = 0x00;		// Zero page as default
	float		ProfileEnd = 0x10000;

	int			RColumnValues[WINDOW_WIDTH];			// Reads
	int			WColumnValues[WINDOW_WIDTH];			// Writes
	int			EColumnValues[WINDOW_WIDTH];			// Executes
	int			RWColumnValues[WINDOW_WIDTH];		// Reads+Writes
	int			RWEColumnValues[WINDOW_WIDTH];		// Reads+Writes+Executes
	int			MaxRColumnValues;
	int			MaxWColumnValues;
	int			MaxEColumnValues;
	int			MaxRWColumnValues;
	int			MaxRWEColumnValues;

	int			ColumnAddress[WINDOW_WIDTH];
	int*		pCurrentColumn;
	int			CurrentMax;

	int			g_ProfileMode = ReadWriteExecute;
	int			g_BarMode = Accumulate;

	int			g_MarkColumnStart = -1;
	int			g_MarkColumnEnd = -1;
	int			g_MouseHover = 0;

	int			g_ProfileUpdateRate=10;
	int			g_ProfileUpdateCurrent=1;
	int			g_ResetProfile = 1;

	float		g_ViewScale = 1.0f;

	HBRUSH		g_Brush;

#define	WHITE		0
#define	RED			1
#define	GREEN		2
#define	BLUE		3
	unsigned int	Cols[]={	0xffffff,0xff0000,0x00ff00,0x0000ff,0xffff00,0xff00ff,0x00ffff,
								0x800080,0xff0080,0x00ff80,0x8000ff,0xffff80,0xff80ff,0x80ffff,
								0x404040,0x804040,0x408030,0x404080,0x808040,0x804080,0x408080
	};

	HGDIOBJ g_OriginalPen;
	HBRUSH	g_OriginalBrush;
	int	g_MouseX,g_MouseY;
	int	MouseLButton;
	int	MouseRButton;

//************************************************************************************
//
//
//
//************************************************************************************
void ResetProfile( void )
{
	for(int i=0;i<65536;i++){
		ReadProfile[i] = 0;
		WriteProfile[i] = 0;
		ExecuteProfile[i] = 0;
	}
	MaxRead = 0;
	MaxWrite = 0;
	MaxExecute = 0;
}


//************************************************************************************
//
//
//
//************************************************************************************
void	SetInk( unsigned _colour )
{
		SetDCPenColor(g_hDC, _colour);
}

//************************************************************************************
//
//
//
//************************************************************************************
int g_nLastX;
int g_nLastY;
void Draw( int x1, int y1, int x2, int y2 )
{
	MoveToEx( g_hDC, x1,y1, NULL );
	LineTo(g_hDC, x2,y2 );
	g_nLastX = x2;
	g_nLastY = y2;
}


//************************************************************************************
//
//
//
//************************************************************************************
extern	char* SymbolLookUp[0x10000];
int LastLMouse=0;
void	DrawMouseInfo( void )
{
	char	Buffer[1024];

//	void Hex (int number,int count)
	
	if(( g_MouseY<WINDOW_HEIGHT-2) && (g_MouseX>11) && (g_MouseX<(WINDOW_WIDTH+11)) )
	{
		int	Column = (g_MouseX-11);
		float	ColumnRange = (float)((int)(WINDOW_WIDTH/(ProfileEnd-ProfileStart)));
		float	ColumnWidth;
		// Width of a single profile column
		if( ColumnRange<1.0f ) ColumnWidth = 1.0f; else ColumnWidth = ColumnRange;

		Column = (int)((float)Column/ColumnWidth);
		Column = (int)(Column); //+ProfileStart);
		
		int LookUpColumn;

		LookUpColumn = Column;
		g_MouseHover = Column;
		if( LastLMouse==0 && MouseLButton==1){
			g_MarkColumnStart = Column;
			g_MarkColumnEnd = Column;
			LastLMouse = 1;
		}if( LastLMouse==1 && MouseLButton==1){
			g_MarkColumnEnd = Column;			
		}if( LastLMouse==1 && MouseLButton==0){
			LastLMouse = 0;
		}

		char* pSym=NULL;
		int add=ColumnAddress[Column];
		while(add>0){
			if(SymbolLookUp[add]!=NULL) {
				pSym=SymbolLookUp[add];
				break;
			}
			add--;
		}

		if(pSym==NULL){
			sprintf(Buffer, "0x%x: %d",ColumnAddress[Column], pCurrentColumn[Column] );
		}else{
			sprintf(Buffer, "0x%x: %d\n%s",ColumnAddress[Column], pCurrentColumn[Column],pSym );
		}

		RECT	Rect;
		Rect.left = g_MouseX;
		Rect.top = g_MouseY;
		Rect.bottom = g_MouseY+24;
		Rect.right = g_MouseX+64;

		//SetInk(0xffeb3f);
		DrawText( g_hDC, Buffer, strlen(Buffer),&Rect,DT_LEFT|DT_NOCLIP);
	}
}

//************************************************************************************
//
//
//
//************************************************************************************
void AccumulateBarData( float ColumnRange )
{
	float blockcount = 0;
	int count=0;
	int address = (int)ProfileStart;
	while(address<(float)ProfileEnd)
	{
		RColumnValues[count] += ReadProfile[address];
		if( MaxRColumnValues<RColumnValues[count]) MaxRColumnValues = RColumnValues[count];

		WColumnValues[count] += WriteProfile[address];
		if( MaxWColumnValues<WColumnValues[count]) MaxWColumnValues = WColumnValues[count];

		EColumnValues[count] += ExecuteProfile[address];
		if( MaxEColumnValues<EColumnValues[count]) MaxEColumnValues = EColumnValues[count];

		RWColumnValues[count] += (ReadProfile[address]+WriteProfile[address]);
		if( MaxRWColumnValues<RWColumnValues[count]) MaxRWColumnValues= RWColumnValues[count];

		RWEColumnValues[count] += (ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address]);
		if( MaxRWEColumnValues<RWEColumnValues[count]) MaxRWEColumnValues= RWEColumnValues[count];

		address++;
		blockcount+= ColumnRange;
		if( blockcount>=1.0f){
			blockcount =0 ;//-= 1.0f;
			count++;
			ColumnAddress[count]=address;
			RColumnValues[count] = 0;
			WColumnValues[count] = 0;
			EColumnValues[count] = 0;
			RWColumnValues[count] = 0;
			RWEColumnValues[count] = 0;
		}
	}
}


//************************************************************************************
//
//
//
//************************************************************************************
void MaxBarData( float ColumnRange )
{
	float blockcount = 0;
	int count=0;
	int address = (int)ProfileStart;
	while(address<(float)ProfileEnd)
	{
		if( RColumnValues[count]<ReadProfile[address]) RColumnValues[count] = ReadProfile[address];
		if( WColumnValues[count]<WriteProfile[address]) WColumnValues[count] = WriteProfile[address];
		if( EColumnValues[count]<ExecuteProfile[address]) EColumnValues[count] = ExecuteProfile[address];
		if( RWColumnValues[count]<(ReadProfile[address]+WriteProfile[address])) RWColumnValues[count] = (ReadProfile[address]+WriteProfile[address]);
		if( RWEColumnValues[count]<(ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address])) RWEColumnValues[count] = (ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address]);

		address++;
		blockcount+= ColumnRange;
		if( blockcount>=1.0f){
			if( MaxRColumnValues<RColumnValues[count]) MaxRColumnValues = RColumnValues[count];
			if( MaxWColumnValues<WColumnValues[count]) MaxWColumnValues = WColumnValues[count];
			if( MaxEColumnValues<EColumnValues[count]) MaxEColumnValues = EColumnValues[count];
			if( MaxRWColumnValues<RWColumnValues[count]) MaxRWColumnValues= RWColumnValues[count];
			if( MaxRWEColumnValues<RWEColumnValues[count]) MaxRWEColumnValues= RWEColumnValues[count];

			blockcount =0 ;//-= 1.0f;
			count++;
			ColumnAddress[count]=address;
			RColumnValues[count] = 0;
			WColumnValues[count] = 0;
			EColumnValues[count] = 0;
			RWColumnValues[count] = 0;
			RWEColumnValues[count] = 0;
		}
	}
}

//************************************************************************************
//
//
//
//************************************************************************************
void MinBarData( float ColumnRange )
{
	float blockcount = 0;
	int count=0;
	int address = (int)ProfileStart;
	RColumnValues[0]= 0x7fffffff;
	WColumnValues[0]= 0x7fffffff;
	EColumnValues[0]= 0x7fffffff;
	RWColumnValues[0]= 0x7fffffff;
	RWEColumnValues[0]= 0x7fffffff;
	while(address<(float)ProfileEnd)
	{
		if( RColumnValues[count]>ReadProfile[address]) RColumnValues[count] = ReadProfile[address];
		if( WColumnValues[count]>WriteProfile[address]) WColumnValues[count] = WriteProfile[address];
		if( EColumnValues[count]>ExecuteProfile[address]) EColumnValues[count] = ExecuteProfile[address];
		if( RWColumnValues[count]>(ReadProfile[address]+WriteProfile[address])) RWColumnValues[count] = (ReadProfile[address]+WriteProfile[address]);
		if( RWEColumnValues[count]>(ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address])) RWEColumnValues[count] = (ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address]);

		address++;
		blockcount+= ColumnRange;
		if( blockcount>=1.0f){
			if( MaxRColumnValues<RColumnValues[count]) MaxRColumnValues = RColumnValues[count];
			if( MaxWColumnValues<WColumnValues[count]) MaxWColumnValues = WColumnValues[count];
			if( MaxEColumnValues<EColumnValues[count]) MaxEColumnValues = EColumnValues[count];
			if( MaxRWColumnValues<RWColumnValues[count]) MaxRWColumnValues= RWColumnValues[count];
			if( MaxRWEColumnValues<RWEColumnValues[count]) MaxRWEColumnValues= RWEColumnValues[count];

			blockcount =0 ;//-= 1.0f;
			count++;
			ColumnAddress[count]=address;
			RColumnValues[count]= 0x7fffffff;
			WColumnValues[count]= 0x7fffffff;
			EColumnValues[count]= 0x7fffffff;
			RWColumnValues[count]= 0x7fffffff;
			RWEColumnValues[count]= 0x7fffffff;
		}
	}
}

//************************************************************************************
//
//
//
//************************************************************************************
void AverageBarData( float ColumnRange )
{
	float blockcount = 0;
	int count=0;
	int address = (int)ProfileStart;
	int	BarCount = 0;
	while(address<(float)ProfileEnd)
	{
		RColumnValues[count] += ReadProfile[address];
		WColumnValues[count] += WriteProfile[address];
		EColumnValues[count] += ExecuteProfile[address];
		RWColumnValues[count] += (ReadProfile[address]+WriteProfile[address]);
		RWEColumnValues[count] += (ReadProfile[address]+WriteProfile[address]+ExecuteProfile[address]);

		address++;
		blockcount+= ColumnRange;
		BarCount++;
		if( blockcount>=1.0f){
			RColumnValues[count] /= BarCount;			// average
			WColumnValues[count] /= BarCount;			// average
			EColumnValues[count] /= BarCount;			// average
			RWColumnValues[count] /= BarCount;			// average
			RWEColumnValues[count] /= BarCount;			// average
			BarCount=0;
			
			if( MaxRColumnValues<RColumnValues[count]) MaxRColumnValues = RColumnValues[count];
			if( MaxWColumnValues<WColumnValues[count]) MaxWColumnValues = WColumnValues[count];
			if( MaxEColumnValues<EColumnValues[count]) MaxEColumnValues = EColumnValues[count];
			if( MaxRWColumnValues<RWColumnValues[count]) MaxRWColumnValues= RWColumnValues[count];
			if( MaxRWEColumnValues<RWEColumnValues[count]) MaxRWEColumnValues= RWEColumnValues[count];

		
			blockcount =0 ;//-= 1.0f;
			count++;
			ColumnAddress[count]=address;
			RColumnValues[count] = 0;
			WColumnValues[count] = 0;
			EColumnValues[count] = 0;
			RWColumnValues[count] = 0;
			RWEColumnValues[count] = 0;
		}
	}
}


//************************************************************************************
//
//
//
//************************************************************************************
void	DrawProfile( HWND hDlg, HDC hDC )
{
	float	width = (float)WINDOW_WIDTH;		// drawing area
	float	height = WINDOW_HEIGHT;				// 

	float	ColumnRange = width/(ProfileEnd-ProfileStart);
	float	ColumnWidth;

	// Width of a single profile column
	if( ColumnRange<1.0f ) ColumnWidth = 1.0f; else ColumnWidth = ColumnRange;


	float BarStep = 32.0f;							// Bar every 32 bytes
	float BarCount = (float)WINDOW_WIDTH/BarStep;	//

	SetInk( 0xffffff );
	int x=(int)BarCount;
	for(int i=0;i<BarStep;i++){
		Draw(x,0, x,WINDOW_HEIGHT);
		x+=(int)BarCount;
	}
	// Draw faint background lines
	SetInk( 0x000000 );
	Draw(0,0, 0,WINDOW_HEIGHT);
	Draw(0,WINDOW_HEIGHT, WINDOW_WIDTH,WINDOW_HEIGHT);
	Draw(WINDOW_WIDTH,WINDOW_HEIGHT,  WINDOW_WIDTH,0);
	Draw(0,0,  WINDOW_WIDTH,0);

	// First calcualte each BAR size by accumulating all the bytes in between
	float dx = ColumnRange;
	int address = (int)ProfileStart;
	float blockcount = 0;
	int count=0;
	RColumnValues[0] = 0;
	WColumnValues[0] = 0;
	EColumnValues[0] = 0;
	RWColumnValues[0] = 0;
	RWEColumnValues[0] = 0;
	MaxRColumnValues = 0;
	MaxWColumnValues = 0;
	MaxEColumnValues = 0;
	MaxRWColumnValues = 0;
	MaxRWEColumnValues = 0;
	ColumnAddress[0]=address;
	
	switch( g_BarMode ){
		case	Accumulate:	AccumulateBarData(ColumnRange); break;
		case	Average:	AverageBarData(ColumnRange); break;
		case	Maximum:	MaxBarData(ColumnRange); break;
		case	Minimum:	MinBarData(ColumnRange); break;
	}
	
	
	
	//
	// Now DRAW the profile data out.
	//

	switch( g_ProfileMode )
	{
		case	ReadOnly:		pCurrentColumn = &(RColumnValues[0]);
								CurrentMax = MaxRColumnValues;
								break;
		case	WriteOnly:		pCurrentColumn = &(WColumnValues[0]);
								CurrentMax = MaxWColumnValues;
								break;
		case	ExecuteOnly:	pCurrentColumn = &(EColumnValues[0]);
								CurrentMax = MaxEColumnValues;
								break;
		case	ReadWrite:		pCurrentColumn = &(RWColumnValues[0]);
								CurrentMax = MaxRWColumnValues;
								break;
		case	ReadWriteExecute:
		default:
								pCurrentColumn = &(RWEColumnValues[0]);
								CurrentMax = MaxRWEColumnValues;
								break;
	}




	int Cols[2]= {0x267cfe, 0x0049b6 };
	x=0;
	float scaler = ((float)WINDOW_HEIGHT/(float)CurrentMax);
	if( g_ViewScale>0.0f){ 
		scaler=g_ViewScale; 
	}



	int	Start = g_MarkColumnStart;
	int	Stop = g_MarkColumnEnd;
	if( Start>Stop ) {
		Stop = g_MarkColumnStart;
		Start = g_MarkColumnEnd;
	}
	for(int i=0;i<WINDOW_WIDTH;i++){
		RECT	Rect;
		Rect.top = WINDOW_HEIGHT-((int)((float)pCurrentColumn[i]*scaler));
		if( Rect.top<0 ) Rect.top=0;

		if( Rect.top ==WINDOW_HEIGHT ) Rect.top = WINDOW_HEIGHT-1;
		Rect.left = x;
		Rect.bottom = WINDOW_HEIGHT;
		Rect.right = Rect.left+(int)ColumnWidth;

		SetDCBrushColor(g_hDC, Cols[i&1]);
		if( i==g_MouseHover){
			SetDCBrushColor(g_hDC, 0x00ff00);
		}else if( i>=Start && i<=Stop ){
			SetDCBrushColor(g_hDC, 0xe67b76);
		}
		FillRect( g_hDC, &Rect,g_Brush);
		x+=(int)ColumnWidth;
	}

	DrawMouseInfo();
}



//************************************************************************************
//
//
//
//************************************************************************************
BOOL DrawProfileDataProc( HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT lpPaint;	 // paint information

	switch (iMsg)
	{
		case WM_PAINT:
			{
				g_hDlg = hDlg;
				BeginPaint( hDlg, &lpPaint );
				{
					g_hDC = GetDC( hDlg );
					g_OriginalPen = (HGDIOBJ)SelectObject(g_hDC,GetStockObject(DC_PEN));
					g_OriginalBrush = (HBRUSH)SelectObject(g_hDC,GetStockObject(DC_BRUSH));
					
					DrawProfile( hDlg, g_hDC );
					
					SelectObject(g_hDC,g_OriginalPen);
					ReleaseDC( hDlg, g_hDC );
				}
				EndPaint( hDlg, &lpPaint );
				return 0;
		}
	}
	return WNProc(hDlg, iMsg, wParam, lParam);
}


//***************************************************************
// 
//***************************************************************
void SetHexMarkers( HWND hDlg)
{
	char str[16];
	sprintf(str,"$%4X", (int)ProfileStart);
	SetDlgItemText( hDlg, IDC_ADD1, str );

	sprintf(str,"$%4X", (int)ProfileEnd);
	SetDlgItemText( hDlg, IDC_ADD3, str );
	
	sprintf(str,"$%4X", (int)(ProfileStart+((ProfileEnd-ProfileStart)/2)));
	SetDlgItemText( hDlg, IDC_ADD2, str );
}

//***************************************************************
// 
//***************************************************************
void	InitDropDowns( HWND hDlg )
{
	char txt[1024];
	sprintf(txt, "Minus4 Profiler:  %s",FileName);
	SetWindowText( hDlg, txt);

	HWND hwndList = GetDlgItem(hDlg, IDC_SCALECOMBO );
	SendMessage(hwndList, CB_RESETCONTENT, 0, 0); 

	LRESULT index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "None"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 1); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Auto"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 2); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "2"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 3); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "4"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 4); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "8"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 5); 

	SendDlgItemMessage(hDlg, IDC_SCALECOMBO, CB_SETCURSEL, 0, 0); 


	//-------------------

	hwndList = GetDlgItem(hDlg, IDC_PROFILECOMBO );
	SendMessage(hwndList, CB_RESETCONTENT, 0, 0); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Read"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 1); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Write"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 2); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Execute"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 3); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Read+Write"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 4); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Read+Write+Execute"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 5); 

	SendDlgItemMessage(hDlg, IDC_PROFILECOMBO, CB_SETCURSEL, 4, 0); 

	// ---------------------------
	hwndList = GetDlgItem(hDlg, IDC_BARCALC );
	SendMessage(hwndList, CB_RESETCONTENT, 0, 0); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Accumulate"); 
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 1); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Average");
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 2); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Maximum");
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 3); 

	index = SendMessage(hwndList, CB_ADDSTRING, 0, (LPARAM) "Minimum");
	SendMessage(hwndList, CB_SETITEMDATA, index, (LPARAM) 4); 

	SendDlgItemMessage(hDlg, IDC_BARCALC, CB_SETCURSEL, 0, 0); 

	// ---------------------------
	//IDC_RESETTIME
	SetDlgItemText(hDlg, IDC_UPDATETIME, "4");

	// --------------------------------
	hwndList = GetDlgItem(hDlg, IDC_RESET ); 
	SendMessage(hwndList, BM_SETCHECK, g_ResetProfile, 1 ); 

	//-----------------------------------
	SetHexMarkers( hDlg );
}




//***************************************************************
// About box code
//***************************************************************

BOOL ProfileDialog ( HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	switch (iMsg)
	{
		case WM_LBUTTONDOWN : {
			MouseLButton = 1;
			break;
		}
		case WM_LBUTTONUP : {
			MouseLButton = 0;
			break;
		}
		case WM_RBUTTONDOWN : {
			MouseRButton = 1;
			break;
		}
		case WM_RBUTTONUP : {
			MouseRButton = 0;
			break;
		}
		case WM_MOUSEMOVE : {
								g_MouseX = LOWORD(lParam);
								g_MouseY = HIWORD(lParam);
								break;
							}

		case WM_INITDIALOG:
			{
				InitDropDowns(hDlg);

				HWND hSlider = GetDlgItem(hDlg, IDC_VIEW );							// get window handle of "frame"
				WNProc = (WNDPROC) GetWindowLong( hSlider, GWL_WNDPROC );		// get old draw proc
				SetWindowLong( hSlider, GWL_WNDPROC, (LONG) DrawProfileDataProc);	// set our new one
				return FALSE;
				break;
			}

		case WM_CLOSE:{
						DestroyWindow (g_ProfileWindow);
						g_ProfileWindow = NULL;
						return FALSE;
					  }

		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
				case IDOK:
					return TRUE;
				case IDCANCEL:
					return FALSE;
				case IDC_RESET:			{
											HWND hwndList = GetDlgItem(hDlg, IDC_RESET ); 
											g_ResetProfile = SendMessage(hwndList, BM_GETCHECK, 0, 0 ); 
											break;
										}
				case IDC_UPDATETIME:		{
											char str[257];
											GetDlgItemText( hDlg, IDC_UPDATETIME, &str[0], 255 );
											g_ProfileUpdateRate = atol(str);
											if( g_ProfileUpdateRate<0 ) g_ProfileUpdateRate=0;
											g_ProfileUpdateCurrent = 1;
											break;
										}
				case	IDC_SCALECOMBO:	{
											switch (HIWORD(wParam)) { 
												case LBN_SELCHANGE:		
														HWND hwndList = GetDlgItem(hDlg, IDC_SCALECOMBO );
														int val = (int) SendMessage(hwndList, CB_GETCURSEL, 0, 0 ); 
														switch( val )
														{
															case	0:	g_ViewScale = 1.0f; break;
															case	1:	g_ViewScale = -1.0f; break;
															case	2:	g_ViewScale = 0.5f;	break;
															case	3:	g_ViewScale = 0.25f; break;
															case	4:	g_ViewScale = 0.125f; break;
															case	5:	g_ViewScale = 1.0f/16.0f; break;
														};
											};
											break;
										}
				case	IDC_PROFILECOMBO:{
											switch (HIWORD(wParam)) { 
												case LBN_SELCHANGE:		
													HWND hwndList = GetDlgItem(hDlg, IDC_PROFILECOMBO );
													int val = (int) SendMessage(hwndList, CB_GETCURSEL, 0, 0 ); 
													switch( val )
													{
														case	0:	g_ProfileMode = ReadOnly; break;
														case	1:	g_ProfileMode = WriteOnly; break;
														case	2:	g_ProfileMode = ExecuteOnly;	break;
														case	3:	g_ProfileMode = ReadWrite;	break;
														case	4:	g_ProfileMode = ReadWriteExecute; break;
													};
											};
											break;
										 }
				case	IDC_BARCALC:{
											switch (HIWORD(wParam)) { 
												case LBN_SELCHANGE:		
													HWND hwndList = GetDlgItem(hDlg, IDC_BARCALC );
													int val = (int) SendMessage(hwndList, CB_GETCURSEL, 0, 0 ); 
													switch( val )
													{
														case	0:	g_BarMode = Accumulate; break;
														case	1:	g_BarMode = Average; break;
														case	2:	g_BarMode = Maximum;	break;
														case	3:	g_BarMode = Minimum;	break;
													};
											};
											break;
										 }
				case IDC_ZPAGE:			ProfileStart = 0x00;
										ProfileEnd = 0x100;
										SetHexMarkers( hDlg );
										break;
				case IDC_64K:			ProfileStart = 0x00;
										ProfileEnd = 0x10000;
										SetHexMarkers( hDlg );
										break;
				case IDC_ZOOM:{
									if( g_MarkColumnEnd!=-1 && g_MarkColumnStart!=-1 )
									{
										int Start = g_MarkColumnStart;
										int Stop = g_MarkColumnEnd;
										if( Start>Stop){
											Start = g_MarkColumnEnd;
											Stop = g_MarkColumnStart;
										}
										Start = ColumnAddress[Start];
										Stop = ColumnAddress[Stop];
										ProfileStart = (float)Start;
										ProfileEnd = (float)Stop;
										g_MarkColumnEnd = -1;
										g_MarkColumnStart = -1;

										SetHexMarkers( hDlg );
									}
							  }
										
			}
			break;
	}
	return FALSE;
}
