//******************************************************************************
//
//
//							    SYMBOL.C
//
//								   By
//
//							   Mike Dailly
//
//						  Copyright 1999,2000
//
// Function
// --------
// This file contains the functions for symbol creation and allocation.
// Symbols are maintained using a binary tree and a first letter hashing table.
// Symbols are all upper case ( I hate case sensative symbols ).
//
//******************************************************************************
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "myheader.h"
//#include <yylex.h>
//#include <lexbuild.h>
//#include <main.h>
#include <symbol.h>
#include <malloc.h>
#include	<io.h>
#include	<sys\stat.h>
#include	<fcntl.h>

#define	BUFFER_SIZE		(1024*1024*10)

PSym		HashingTable[256];		// symbol hashing table  '@' is local
PSymSpace	pFirst = NULL;
PSymSpace	pCurrent = NULL;
uint32		dwSymbolCount=0;
uint32		dwSymbolMemoryUsed=0;
uint32		dwLocalSymbolCount;
uint32		SymbolSpace=0;

uint8*		g_pOutputBuffer;
uint32		g_OutputIndex;
int			g_OutputHandle;
int			g_OutputCount;

char		pszLastLabel[4096];			// pointer to LAST normal label (used to append locals)
PSym		pLastSearch;			// Last location (before symbol links lost)
uint32		*pLastParent;			// Last parent pointer (actual DWORD value)
int32		LastSymCompare;			// Last symbol search compare
PSym		pLastSymbol;			// Last add symbol location (even if failed due to redefinition)
//int32		g_InMacro=0;			// if being defined inside a macro, then this == 1
int32		g_MacroCounter=0;		// auto inc counter to allow local labels inside macros
char		g_LastLastSymbol[4096]={"??START??"};		// previous symbol before this one....

//*******************************************************************************
// Free all symbol blocks
//*******************************************************************************
void FreeSymbols( void )
{
	PSymSpace	pSyms = pFirst;
	PSymSpace	pNext;

	while(pSyms!=NULL){
		pNext = pSyms->pNext;
		free( pSyms );
		pSyms = pNext;
	}
}



//*******************************************************************************
//
//	Init Symbol system.
//
//*******************************************************************************
void	InitSymbols( void )
{
	SymbolSpace = SYMBOL_BLOCK_COUNT;						// space for 16384 symbols
	pFirst = (PSymSpace) malloc( SYMBOL_BLOCK_COUNT + sizeof(tSymSpace) );	// allocate space...
	//dwSymbolMemoryUsed+=SYMBOL_BLOCK_COUNT + sizeof(tSymSpace);
	pFirst->pNext = NULL;
	pFirst->dwSize = sizeof(tSymSpace);
	pFirst->pMemory = ((uint8*)pFirst)+sizeof(tSymSpace);
	pCurrent = pFirst;

	memset( HashingTable,0, sizeof(PSym)*256 );
	dwSymbolCount=0;
	dwSymbolMemoryUsed=0;
	dwLocalSymbolCount=0;
	SymbolSpace=0;
	g_MacroCounter=0;
}

//*******************************************************************************
//
//	Name:		AllocSpace
//	Function:	Allocate memory for a symbol. If the current "block" has run
//				out of space, create a new block and link with old. Then
//				allocate from NEW block
//	
//	In:			None
//	Out:		PSym pointer
//
//*******************************************************************************
void*	AllocSpace( unsigned int size )
{
	PSym	pSym;

	if( SymbolSpace < size ){
		PSymSpace	pNext;

		pNext = (PSymSpace) malloc( SYMBOL_BLOCK_COUNT + sizeof(tSymSpace) );	// allocate space...
		pCurrent->pNext = pNext;
		pCurrent = pNext;

		SymbolSpace       = SYMBOL_BLOCK_COUNT;						// space for 16384 symbols
		pCurrent->pNext   = NULL;
		pCurrent->dwSize  = sizeof(tSymSpace);
		pCurrent->pMemory = ((uint8*)pFirst)+sizeof(tSymSpace);
	}

	pSym = (PSym) pCurrent->pMemory;
	if( size == 0 ){
		size = sizeof(tSym);
	}

	pCurrent->pMemory+=size;
	pCurrent->dwSize += size;		// keep count of memory used so far...
	SymbolSpace -= size;
	dwSymbolMemoryUsed+=size;
	return pSym;
}


//*******************************************************************************
//
//	Name:		FindSymbol
//	Function:	Search our binary tree for the symbol we want.
//				The main core is using the strcmp() call and using the
//				< 0  and >0 flags. to speed this up I store the string length
//				as well so we can recurse past totaly different symbols.
//
//	In:			pszName = string to search for.
//	Out:		NULL = not found,
//				PSym = location of symbol
//
//				pLastSearch    = last valid location of search.
//				LastSymCompare = last compare of last search.
//
//*******************************************************************************
PSym FindSymbol( char *pszName )
{
	PSym	pSym;
	int32	i;
	uint32	Length;
	//	uint32	SymLen;
	char	szLocal[SYMBOL_LEN];
	char	*pszLocal;
	char	tmp[33];

//	if( strcmp(pLex->m_yyString,"B")==0){
//		_asm int 3;
//	}


	if(( pszName[0] == '@' ) || ( pszName[0] == '!' )){
		strcpy( &szLocal[0], pszLastLabel);			// we could prepend a counter here for macros...
		strcat( &szLocal[0], pszName );

		if( g_MacroCounter!=0 ) {
			_itoa( g_MacroCounter, tmp, 10) ;
			strcat( &szLocal[0], tmp );
		}
		pszLocal = &szLocal[0];
	}
	else{
		pszLocal = pszName;
	}

	pSym = HashingTable[ pszLocal[0] ];
	pLastParent = (uint32*) &(HashingTable[ pszLocal[0] ]);
	Length = strlen( pszName )+1;

	pLastSearch = NULL;
	LastSymCompare = 0;
	if( pSym == NULL ) return NULL;			// can't find start...


	pLastSearch = pSym;
	while( pSym!=NULL ){
		//SymLen = pSym->dwFlags&0xff;		// get string length
		//if( Length == SymLen ){
		i = strcmp( pszLocal, &pSym->Name[0] );
		//}
		//else{
		//	i = Length - SymLen;			// simulate a <0 and >0 branch....
		//}

		LastSymCompare = i;
		if( i == 0 ) return pSym;			// exact match found!!

		if( i<0 ){
			pLastSearch = pSym;
			pLastParent = (uint32*) &pLastSearch->pLeft;
			pSym = pSym->pLeft;
		}
		else{
			pLastSearch = pSym;
			pLastParent = (uint32*) &pLastSearch->pRight;
			pSym = pSym->pRight;
		}
	}


	return NULL;
}

//*******************************************************************************
//
//	Name:		AddSymbol
//	Function:	Add a symbol to our symbol list.
//
//	In:			pszName = symbol to add (uppercase)
//				Value   = value to assign to symbol
//				flags   = flags to store with symbol
//	Out:		PSym = NULL for already assigned
//				PSym = pointer to symbol				
//
//*******************************************************************************
PSym	AddSymbol( char *pszName, int32 Value, int32 Value2, uint32 flags )
{
	PSym	pSpot;
	PSym	pSymbol;
	uint32	dSize;
	char	szLocal[SYMBOL_LEN];
	char	*pszLocal;
	int32	sdIndex;
	char	tmp[SYMBOL_LEN];


//	if( strcmp(pLex->m_yyString,"XPLY")==0){
//		_asm int 3;
//	}

	dSize = strlen( pszName )+1;
	if( dSize >SYMBOL_LEN ) pszName[SYMBOL_LEN+1]=0x00;				// clamp symbol;


	if(( pszName[0] == '@' ) || ( pszName[0] == '!' )){
		strcpy( &szLocal[0], pszLastLabel);			// we could prepend a counter here for macros...
		strcat( &szLocal[0], pszName );

		if( g_MacroCounter!=0 ) {
			_itoa( g_MacroCounter, tmp, 10) ;
			strcat( &szLocal[0], tmp );
		}
		pszLocal = &szLocal[0];
		dwLocalSymbolCount++;
	}
	else{
		pszLocal = pszName;
	}



	// First see if we have ANY entries here.....
	sdIndex = pszLocal[0];
	if( HashingTable[ sdIndex ] == NULL ){
		pSymbol = (PSym)AllocSpace(0);
		HashingTable[ sdIndex ] = pSymbol;
		dwSymbolCount++;
	}
	else{
		pSpot = FindSymbol( pszLocal );				// find symbol spot
		pLastSymbol = pSpot;
		if( pSpot != NULL )	return NULL;			// *** return error, symbol defined ***
		pSymbol = (PSym)AllocSpace(0);
		if( LastSymCompare < 0 ){
			pLastSearch->pLeft = pSymbol;
		}
		else{
			pLastSearch->pRight = pSymbol;
		}
		dwSymbolCount++;
	}


	// Fill in symbol info.
	pLastSymbol = pSymbol;
	pSymbol->pLeft   = NULL;
	pSymbol->pRight  = NULL;
	pSymbol->dwFlags = flags | (strlen( pszLocal )+1);
	pSymbol->Value   = Value;
	strcpy( &pSymbol->Name[0], pszLocal );

	if(( pszName[0] != '@' ) && ( pszName[0] != '!' )){
		if( pszLastLabel!=NULL ){
			strcpy(g_LastLastSymbol,pszLastLabel);
		}else{
			g_LastLastSymbol[0]=0x00;
		}
		strcpy(pszLastLabel,pSymbol->Name);
		//pszLastLabel = &pSymbol->Name[0];
	}


	return pSymbol;
}


//*******************************************************************************
//
//	Name:		AddSymbol
//	Function:	Add a symbol to our symbol list.
//
//	In:			pSym = already created node to add into symbol tables binary-tree
//	Out:		PSym = NULL for already assigned
//				PSym = pointer to symbol				
//
//*******************************************************************************
PSym AddNode( PSym pSym)
{
	PSym	pSpot;
	PSym	pSymbol;
	char	*pszLocal;
	int32	sdIndex;

	pszLocal = pSym->Name;


	// First see if we have ANY entries here.....
	sdIndex = pszLocal[0];
	if( HashingTable[ sdIndex ] == NULL ){
		pSymbol = (PSym)pSym;
		HashingTable[ sdIndex ] = pSymbol;
	}
	else{
		pSpot = FindSymbol( pszLocal );				// find symbol spot
		pLastSymbol = pSpot;
		if( pSpot != NULL )	return NULL;			// *** return error, symbol defined ***
		pSymbol = pSym;
		if( LastSymCompare < 0 ){
			pLastSearch->pLeft = pSymbol;
		}
		else{
			pLastSearch->pRight = pSymbol;
		}
	}


	// Fill in symbol info.
	pLastSymbol = pSymbol;
	pSymbol->pLeft   = NULL;
	pSymbol->pRight  = NULL;
	return pSymbol;
}

//*******************************************************************************
//
//	Name:		AddRecursively
//	Function:	Takes a binary tree and adds it into another binary tree
//
//	In:			pSym = root node
//	Out:		none
//
//*******************************************************************************
void	AddRecursively( PSym pSym )
{
	PSym pLeft;
	PSym pRight;

	if( pSym==NULL ) return;
	pLeft  = pSym->pLeft;
	pRight = pSym->pRight;

	AddNode( pSym );
	if( pLeft!=NULL ) AddRecursively( pLeft );
	if( pRight!=NULL ) AddRecursively( pRight );
}


//*******************************************************************************
//
//	Name:		DumpSymbolTable
//	Function:	Save symbol table out to disk
//
//	In:			pOutputName = Name to save as (extension is wrong)
//	Out:		none
//
//*******************************************************************************
void	DumpSyms( PSym pSymbol )
{
	uint32*	p32Data;

	if( pSymbol==NULL ) return;
	if( pSymbol->pLeft!=NULL ) DumpSyms(pSymbol->pLeft );

	p32Data = (uint32*) &(g_pOutputBuffer[g_OutputIndex]);
	*p32Data = pSymbol->Value;
	g_OutputIndex+=4;

	// Copy string over
	strcpy( (char*) &(g_pOutputBuffer[g_OutputIndex]), pSymbol->Name );
	g_OutputIndex += (strlen( pSymbol->Name )+1);

	g_OutputCount++;

	if( pSymbol->pRight!=NULL ) DumpSyms(pSymbol->pRight );
}

//*******************************************************************************
//
//	Name:		DumpSymbolTable
//	Function:	Save symbol table out to disk
//
//	In:			pOutputName = Name to save as (extension is wrong)
//	Out:		none
//
//*******************************************************************************
void DumpSymbolTable( char* pOutputName )
{
	char	Name[1024];
	uint32*	p32Data;

	strcpy(Name,pOutputName);
	strcat(Name,".sym");
	g_pOutputBuffer = (uint8*)malloc( BUFFER_SIZE );
	g_OutputIndex = 0;
	g_OutputCount=0;

	g_OutputHandle=_open(Name, O_RDWR|O_CREAT|_O_TRUNC|O_BINARY,_S_IREAD | _S_IWRITE);
    g_OutputIndex=4;
	for(int i=0;i<256;i++){
		DumpSyms( HashingTable[i] );		// send parent into recursive loop
	}

	p32Data = (uint32*) &(g_pOutputBuffer[0]);
	*p32Data = g_OutputCount;
	int BytesSaved=_write(g_OutputHandle, g_pOutputBuffer, g_OutputIndex );
	

	_close( g_OutputHandle );
}


//*******************************************************************************
//
//	Name:		FindValue2
//	Function:	Search for a symbol with a specific value
//
//	In:			pSym	= symbol to look at
//				value	= value to find
//	Out:		none
//
//*******************************************************************************
PSym	FindValue2( PSym pSym,int value )
{
	PSym	pRet;
	
	if( pSym->pLeft!=NULL ){
		pRet = FindValue2(pSym->pLeft, value);
		if( pRet!=NULL ) {
			return pRet;
		}
	}

	if( pSym->Value == value ) return pSym;

	if( pSym->pRight!=NULL ) {
		pRet = FindValue2(pSym->pRight, value);
		if( pRet!=NULL ) {
			return pRet;
		}
	}
	return NULL;
}
//*******************************************************************************
//
//	Name:		FindValue
//	Function:	Search for a symbol with a specific value
//
//	In:			value	= value to find
//	Out:		char* = found symbol name...or NULL
//
//*******************************************************************************
char*	FindValue( int value )
{
	PSym pSym;
	for(int i=0;i<256;i++)
	{
		if( HashingTable[i]!=NULL ){
			pSym = FindValue2( HashingTable[i], value );		// send parent into recursive loop
			if( pSym!=NULL ) {
				return &(pSym->Name[0]);
			}
		}
	}
	return NULL;
}


